from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from pptx import Presentation # python-pptx

if TYPE_CHECKING:
	from typing import Iterator, List

def read_ppt(path):
	# type: (str, ) -> Iterator[List[str]]

	prs = Presentation(path)
	for slide in prs.slides:
		shapes = [shape.text for shape in slide.shapes if hasattr(shape, "text")]
		yield shapes



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import PY36_PLUS
from .os import PathLike

try:
	from pathlib import Path, PurePath  # noqa: 401

	if not PY36_PLUS:
		PurePath.__fspath__ = PurePath.__str__

except ImportError:
	from pathlib2 import Path, PurePath  # noqa: 401

assert hasattr(PurePath, "__fspath__")
PathLike.register(PurePath)



from __future__ import absolute_import, division, print_function, unicode_literals

import sys

if sys.version_info >= (3, 0):
	from abc import ABC
else:
	from abc import ABCMeta
	ABC = ABCMeta(b"ABC", (object,), {"__slots__": ()})



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map, zip
from future.utils import PY2, viewkeys, viewitems

import re
from itertools import chain
from locale import strxfrm
from functools import partial
from collections import OrderedDict
from typing import TYPE_CHECKING

from .iter import switched_enumerate
from .binary import encode_binary, decode_binary

if TYPE_CHECKING:
	from typing import Any, Dict, Iterable, Tuple, Callable, Optional, Sequence, TypeVar, Union, Mapping
	T = TypeVar("T")

english_consonants = "bcdfghjklmnpqrstvwxz" # y?
english_vowels = "aeiouy" # y?

german_consonants = "bcdfghjklmnpqrstvwxyz"
german_vowels = "aeiou"

#def cycleleft(s, start=0, end=-1): # bad
#	return s[:start] + s[start+1:end] + s[start] + s[end+1:]

#def cycleright(s, start=0, end=-1): # bad
#	return s[:start] + s[end] + s[start:end] + s[end+1:]

def toint(obj, default=None):
	# type: (Any, Optional[T]) -> Union[int, Optional[T]]

	""" Converts `obj` to int if possible and returns `default` otherwise.
	"""

	try:
		return int(obj)
	except (ValueError, TypeError):
		return default

def tryint(obj):
	# type: (T, ) -> Union[int, T]

	""" Converts `obj` to int if possible and returns the original object otherwise.
	"""

	try:
		return int(obj)
	except (ValueError, TypeError):
		return obj

def removeprefix(s, prefix):
	# type: (str, str) -> str

	""" If string `s` starts with `pre` cut it off and return the remaining string. """

	if s.startswith(prefix):
		return s[len(prefix):]
	else:
		return s

def removesuffix(s, suffix):
	# type: (str, str) -> str

	""" If string `s` ends with `suffix` cut it off and return the preceding string. """

	if s.endswith(suffix):
		return s[:-len(suffix)]
	else:
		return s

def encode_case(s):
	# type: (str, ) -> bytes

	""" Encode case information of a string to a bit-encoded bytes string.
	"""

	return encode_binary(c.isupper() for c in s)

def decode_case(s, key):
	# type: (str, bytes) -> str

	""" Apply the case information encoded in `key` to string `s`.
	"""

	return "".join(c.upper() if b else c for c, b in zip(s, decode_binary(key)))

def locale_sorted(seq, case_insensitive=True, lower_before_upper=True):
	if case_insensitive:
		if lower_before_upper:
			key = lambda s: (strxfrm(s.lower()), strxfrm(s.swapcase()))
		else:
			key = lambda s: (strxfrm(s.lower()), strxfrm(s))
	else:
		if lower_before_upper:
			key = lambda s: (strxfrm(s.swapcase()), strxfrm(s))
		else:
			key = strxfrm

	return sorted(seq, key=key)

def build_multiple_replace(d, escape=True):
	# type: (Mapping[str, str], bool) -> Callable[[str], str]

	""" Returns a callable, which when applied to a string,
		replaces all the keys in `d` with the corresponding values.
		The replacement happens in iteration order of the mapping.
		If the order of replacement is important, an OrderedDict should
		be used instead of a dict for Python < 3.7.
		The complexity then is linear in  the length of the input string.
	"""

	if escape:
		it = map(re.escape, viewkeys(d))
	else:
		it = viewkeys(d)

	cp = re.compile("(" + "|".join(it) + ")")
	return partial(cp.sub, lambda m: d[m.group(0)])

def multiple_replace(d, s):
	# type: (Mapping[str, str], str) -> str

	""" Uses mapping `d` to replace keys with values in `s`.
		This is a wrapper for `build_multiple_replace`,
		which should be used directly for improved performance
		in case the same `d` is used for multiple `s`.
	"""

	f = build_multiple_replace(d)
	return f(s)

def replace_multiple(s, d): # cython candidate...
	# type: (Iterable[str], Mapping[str, str]) -> str

	""" Use dictionary `d` to replace instances of key with value.
		If input `s` is a string, the keys must be strings of length 1.
		If the input is an iterable of strings,
		the strings will be compared to the keys for equality.
		The complexity is linear in the length of the iterable/string.
		For string inputs and keys of length > 1 see `multiple_replace`.
	"""

	return "".join(d.get(c, c) for c in s)

def backslash_unescape(s):
	# type: (str, ) -> str

	""" Converts strings with backslash-escaped entities like \n or \u1234
		to a string with these entities.
		Example: backslash_unescape("\\n") -> "\n"
	"""

	# for some reason unicode_escape is based on latin-1
	# ascii also works, but the intermittent string would be longer
	# because more things will be backslash-escaped
	return s.encode("latin-1", "backslashreplace").decode("unicode_escape")

_backslashquote_escape = build_multiple_replace(OrderedDict([
	("\\", "\\\\"),
	("\"", "\\\"")
]))
def backslashquote_escape(s):
	# type: (str, ) -> str

	""" Converts \ to \\ and " to \"
	"""

	return _backslashquote_escape(s)

_backslashquote_unescape = build_multiple_replace(OrderedDict([
	("\\\"", "\""),
	("\\\\", "\\")
]))
def backslashquote_unescape(s):
	# type: (str, ) -> str

	""" Unescapes \\ and \" in `s`.
	"""

	return _backslashquote_unescape(s)

_backslashcontrol_escape = build_multiple_replace(OrderedDict([
	("\\", "\\\\"),
	("\t" ,"\\t"),
	("\n", "\\n"),
	("\r", "\\r")
]))
def backslashcontrol_escape(s):
	# type: (str, ) -> str

	return _backslashcontrol_escape(s)

_backslashcontrol_unescape = build_multiple_replace(OrderedDict([
	("\\\\", "\\"),
	("\\t", "\t"),
	("\\n", "\n"),
	("\\r", "\r")
]))
def backslashcontrol_unescape(s):
	# type: (str, ) -> str

	return _backslashcontrol_unescape(s)

def are_parentheses_matched(s, open="([{", close=")]}"):
	# type: (str, str, str) -> bool

	stack = list()
	parentheses = dict(chain(switched_enumerate(open), switched_enumerate(close)))
	assert len(parentheses) == len(open) + len(close)

	for char in s:
		if char in open:
			stack.append(parentheses[char])
		if char in close:
			try:
				if stack.pop() != parentheses[char]: # wrong nesting
					return False
			except IndexError: # stack empty
				return False

	return len(stack) == 0 # unclosed parentheses left

def filter_join(s, it, func=None):
	# type: (str, Iterable[str], Optional[Callable]) -> str

	if func is None:
		return s.join(i for i in it if i)
	else:
		return s.join(i for i in it if func(i))

def surrounding_join(j, it, l="", r=""):
	# type: (str, Iterable[str], str, str) -> str

	""" Example: surrounding_join(", ", ("a", "b", "c"), l="[", r="]") -> "[a], [b], [c]"
	"""

	s = (r + j + l).join(it)
	if s != "":
		return l + s + r
	return ""

def replace_pairs_bytes(s, items):
	# type: (bytes, Dict[bytes, Optional[bytes]]) -> bytes

	frm = b"".join(k for k, v in viewitems(items) if v)
	to = b"".join(v for k, v in viewitems(items) if v)
	delete = b"".join(k for k, v in viewitems(items) if v is None)

	if PY2:
		import string
		table = string.maketrans(frm, to)
	else:
		table = s.maketrans(frm, to)

	return s.translate(table, delete)

def replace_pairs_chars(s, items):
	# type: (str, Dict[str, Optional[str]]) -> str

	if PY2:
		# table = s.maketrans(items) # 'unicode' object has no attribute 'maketrans'
		table = {ord(k):v for k, v in viewitems(items)}
	else:
		table = s.maketrans(items)

	return s.translate(table)

_contains_digit = re.compile(r"\d")
def contains_digit(s):
	# type: (str, ) -> bool

	""" Tests if a digit is contained in string `s`.
	"""

	return _contains_digit.search(s) is not None



import requests

class RasaRestWebhook(object):

	scheme = "http"

	def __init__(self, sender, netloc="localhost:5005", timeout=60):
		# type: (str, str, int) -> None

		self.sender = sender
		self.netloc = netloc
		self.timeout = timeout

	def health(self):
		# type: () -> dict

		url = self.scheme + "://" + self.netloc + "/webhooks/rest/"

		r = requests.get(url, timeout=self.timeout)
		r.raise_for_status()
		return r.json()

	def send_message(self, message):
		# type: (str, ) -> dict

		url = self.scheme + "://" + self.netloc + "/webhooks/rest/webhook"

		r = requests.post(url, json={
			"sender": self.sender,
			"message": message,
		}, timeout=self.timeout)
		r.raise_for_status()
		return r.json()



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import str, bytes

import os.path
from tempfile import mkstemp
from os import remove
from typing import TYPE_CHECKING

from .file import copen
from .compat.os import replace, fspath

if TYPE_CHECKING:
	from typing import Union, Optional, IO
	from .compat.os import PathLike
	PathType = Union[str, PathLike]

# http://stupidpythonideas.blogspot.tw/2014/07/getting-atomic-writes-right.html
class TransactionalCreateFile(object):

	def __init__(self, path, mode="wb", encoding=None, errors=None, newline=None, prefix="tmp"):
		# type: (PathType, str, Optional[str], Optional[str], Optional[str], str) -> None

		is_text = "t" in mode

		self.path = fspath(path)
		suffix = os.path.splitext(self.path)[1].lower()
		curdir = os.path.dirname(self.path)
		fd, self.tmppath = mkstemp(suffix, prefix, curdir, is_text)
		self.fp = copen(fd, mode, encoding=encoding, errors=errors, newline=newline, ext=suffix)

	def commit(self):
		# type: () -> None

		self.fp.close()
		replace(self.tmppath, self.path) # should be atomic

	def rollback(self):
		# type: () -> None

		self.fp.close()
		remove(self.tmppath)

	def __enter__(self):
		# type: () -> IO

		return self.fp

	def __exit__(self, exc_type, exc_value, traceback):
		# at this point the original file is unmodified and the new file exists as tempfile on disk (or in buffer on windows)
		if exc_type:
			self.rollback()
		else:
			self.commit()

def write_file(data, path, mode="wb", encoding=None, errors=None, newline=None):
	# type: (Union[str, bytes], PathType, str, Optional[str], Optional[str], Optional[str]) -> None

	""" Writes/overwrites files in a safe way. That means either the original file
		will be left untouched, or it will be replaced with the complete new file.
	"""

	with TransactionalCreateFile(path, mode, encoding=encoding, errors=errors, newline=newline) as fw:
		fw.write(data)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewkeys
import logging, json, csv
from itertools import islice
from functools import partial, wraps
from datetime import timedelta
from typing import TYPE_CHECKING

from .atomic import TransactionalCreateFile
from .compat import FileNotFoundError
from .datetime import now
from .file import copen
from .filesystem import mdatetime
from .object import args_to_key

if TYPE_CHECKING:
	from typing import Any, Callable, Dict, Optional, TextIO, Union, Iterator, Sequence, Tuple
	from .compat.pathlib import Path
	JsonDict = Dict[str, Any]

if __debug__:
	import jsonschema

logger = logging.getLogger(__name__)

class BuiltinEncoder(json.JSONEncoder):

	def default(self, obj):
		from datetime import date, timedelta
		from uuid import UUID
		from base64 import b85encode

		if isinstance(obj, (set, frozenset)):
			return tuple(obj)
		elif isinstance(obj, complex):
			return [obj.real, obj.imag]
		elif isinstance(obj, date):
			return obj.isoformat()
		elif isinstance(obj, timedelta):
			return obj.total_seconds()
		elif isinstance(obj, UUID):
			return str(obj)
		elif isinstance(obj, bytes):
			return b85encode(obj).decode("ascii") # b85encode doesn't use ", ' or \

		return json.JSONEncoder.default(self, obj)

def read_json_schema(path):
	# type: (str, ) -> JsonDict

	with open(path, "r", encoding="utf-8") as fr:
		return json.load(fr)

def read_json(path, schema=None, cls=None, object_hook=None):
	# type: (str, Optional[Union[str, JsonDict]]) -> Any

	""" Read the json file at `path` and optionally validates the input according to `schema`.
		The validation requires `jsonschema`.
		`schema` can either be a path as well, or a Python dict which represents the schema.
		`cls` and `object_hook` is passed through to `json.load`.
	"""

	with copen(path, "rt", encoding="utf-8") as fr:
		obj = json.load(fr, cls=cls, object_hook=object_hook)

	if schema is None:
		return obj

	from jsonschema import validate

	if isinstance(schema, str):
		schema = read_json_schema(schema)

	validate(obj, schema)
	return obj

def write_json(obj, path, schema=None, ensure_ascii=False, cls=None, indent=None, sort_keys=False, default=None, safe=False):
	# type: (Any, str, Optional[Union[str, JsonDict]], bool, Optional[str], bool, Optional[Callable]) -> None

	""" Writes python object `obj` to `path` as json files and optionally validates the object
		according to `schema`. The validation requires `jsonschema`.
		The remaining optional parameters are passed through to `json.dump`.
		`safe`: if True, don't overwrite original file in case any error occurs
	"""

	if schema:
		from jsonschema import validate

		if isinstance(schema, str):
			schema = read_json_schema(schema)

		validate(obj, schema)

	if safe:
		context = TransactionalCreateFile
	else:
		context = copen

	with context(path, "wt", encoding="utf-8") as fw:
		json.dump(obj, fw, ensure_ascii=ensure_ascii, cls=cls, indent=indent, sort_keys=sort_keys, default=default)

class json_lines(object):

	""" Read and write files in the JSON Lines format (http://jsonlines.org).
	"""

	def __init__(self,
		stream, doclose, cls=None, object_hook=None, parse_float=None,
		parse_int=None, parse_constant=None, object_pairs_hook=None, **kw
	):
		# type: (TextIO, bool, Optional[json.JSONEncoder], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], **Any) -> None

		""" Don't use directly. Use `from_path` or `from_stream` classmethods instead. """

		""" fixme: how should `close` be handled?
			1: If you don't want to close `stream`, just don't call `close()` or use as context manager.
			2: use `doclose` argument to decide
		"""

		self.f = stream
		self.doclose = doclose
		self.newline = "\n"

		self.json_kwargs = {
			"cls": cls,
			"object_hook": object_hook,
			"parse_float": parse_float,
			"parse_int": parse_int,
			"parse_constant": parse_constant,
			"object_pairs_hook": object_pairs_hook,
		}
		self.json_kwargs.update(kw)

	@staticmethod
	def from_path(
		file, mode="rt", encoding="utf-8", errors="strict", newline=None,
		cls=None, object_hook=None, parse_float=None, parse_int=None, parse_constant=None, object_pairs_hook=None, **kw
	):
		# type: (str, str, int, str, str, Optional[str], bool, Optional[Callable], Optional[json.JSONEncoder], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], **Any) -> json_lines

		stream = copen(file, mode, encoding=encoding, errors=errors, newline=newline)
		return json_lines(stream, True, cls, object_hook, parse_float, parse_int, parse_constant, object_pairs_hook, **kw)

	@staticmethod
	def from_stream(
		stream,
		cls=None, object_hook=None, parse_float=None, parse_int=None, parse_constant=None, object_pairs_hook=None, **kw
	):
		# type: (TextIO, Optional[json.JSONEncoder], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], Optional[Callable], **Any) -> json_lines

		return json_lines(stream, False, cls, object_hook, parse_float, parse_int, parse_constant, object_pairs_hook, **kw)

	def __enter__(self):
		# type: () -> json_lines

		return self

	def __exit__(self, exc_type, exc_value, traceback):
		self.close()

	def iterrange(self, start=0, stop=None):
		# type: (int, Optional[int]) -> Iterator

		linenum = start+1
		try:
			for line in islice(self.f, start, stop):
				line = line.rstrip().lstrip("\x00") # fixme: strip \0 is only a temp fix!
				if line:
					yield json.loads(line, **self.json_kwargs)
				linenum += 1

		except ValueError as e: # json.JSONDecodeError in python 3.5+
			e.lineno = linenum
			logger.error("JSON Lines parse error in line %s: '%r'", linenum, line)
			raise

	def __iter__(self):
		# type: () -> Iterator

		return self.iterrange()

	def write(self, obj, skipkeys=False, ensure_ascii=False, check_circular=True, allow_nan=True,
		cls=None, separators=None, default=None, sort_keys=False, **kw):
		# type: (Any, bool, bool, bool, bool, Optional[Callable], Optional[Tuple[str, str]], Optional[Callable], bool, **Any) -> None

		json.dump(obj, self.f, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
			allow_nan=allow_nan, cls=cls, indent=None, separators=separators, default=default, sort_keys=sort_keys,
			**kw)
		self.f.write(self.newline)

	def close(self):
		# type: () -> None

		if self.doclose:
			self.f.close()

def read_json_lines(file, object_hook=None):

	""" Iterate over a JSON Lines `file` object by object.
		`object_hook` is passed through to `json.load`.
	"""

	with json_lines.from_path(file, mode="rt", object_hook=object_hook) as fr:
		for obj in fr:
			yield obj

def jl_to_csv(jlpath, csvpath, keyfunc, mode="xt"):
	# type: (str, str, Callable[[JsonDict], Sequence[str]], str, str) -> None

	with json_lines.from_path(jlpath, "rt") as fr:
		with open(csvpath, mode, encoding="utf-8", newline="") as csvfile:
			fw = csv.writer(csvfile)
			for obj in fr:
				fw.writerow(keyfunc(obj))

def key_to_hash(key, default=None):
	from hashlib import md5
	binary = json.dumps(key, default=default).encode("utf-8")
	return md5(binary).hexdigest()  # nosec

def cache(path, duration=None, ensure_ascii=False, indent=None, sort_keys=False, default=None, object_hook=None):
	# type: (Path, Optional[timedelta], bool, Optional[str], bool, Optional[Callable], Optional[Callable]) -> Callable

	""" Decorator to cache results of function to json files at `path` for `duration`.
		The remaining parameters are passed through to `json.dump`.
	"""

	duration = duration or timedelta.max

	def decorator(func):
		# type: (Callable, ) -> Callable

		@wraps(func)
		def inner(*args, **kwargs):

			hash = key_to_hash(args_to_key(args, kwargs, {}), default=default)
			fullpath = path / hash

			try:
				invalid = now() - mdatetime(fullpath) > duration
			except FileNotFoundError:
				invalid = True

			if invalid:
				path.mkdir(parents=True, exist_ok=True)
				ret = func(*args, **kwargs)
				write_json(ret, fullpath, ensure_ascii=ensure_ascii, indent=indent, sort_keys=sort_keys, default=default, safe=True)
				return ret
			else:
				return read_json(fullpath, object_hook=object_hook)

		return inner
	return decorator

def jsonlines_cache(path, duration=None, ensure_ascii=False, sort_keys=False, default=None, object_hook=None):
	# type: (Path, Optional[timedelta], bool, Optional[str], bool, Optional[Callable], Optional[Callable]) -> Callable

	""" Decorator to cache results of function to jsonlines files at `path` for `duration`.
		The remaining parameters are passed through to `json_lines.from_path`.
	"""

	duration = duration or timedelta.max

	def decorator(func):
		# type: (Callable, ) -> Callable

		@wraps(func)
		def inner(*args, **kwargs):

			hash = key_to_hash(args_to_key(args, kwargs, {}), default=default)
			fullpath = path / hash

			try:
				invalid = now() - mdatetime(fullpath) > duration
			except FileNotFoundError:
				invalid = True

			if invalid:
				path.mkdir(parents=True, exist_ok=True)
				with TransactionalCreateFile(fullpath, "wt") as stream:
					with json_lines.from_stream(stream, ensure_ascii=ensure_ascii, sort_keys=sort_keys, default=default) as fw:
						# if `func` raises, TransactionalCreateFile makes sure that the original
						# cache file remains unmodified and that temporary files are deleted
						for obj in func(*args, **kwargs):
							fw.write(obj)
							yield obj
			else:
				with json_lines.from_path(fullpath, "rt", object_hook=object_hook) as fr:
					for obj in fr:
						yield obj

		return inner
	return decorator

class JsonLinesFormatter(logging.Formatter):

	""" A JSON Lines formatter for the Python logging library.
		It expects a `dict` as logging message.
		For example:
			logger = logging.getLogger("jsonlines-test")
			handler = logging.StreamHandler()
			formatter = JsonLinesFormatter()
			handler.setFormatter(formatter)
			logger.addHandler(handler)
			logger.warning({"msg": "Hello world!", "level": "greeting"})
	"""

	myfields = {
		"datetime": now,
	}

	def __init__(self, include=frozenset(), default=None):
		logging.Formatter.__init__(self)
		self.include_b = include & viewkeys(self.myfields)
		self.dumps = partial(json.dumps, ensure_ascii=False, indent=None, sort_keys=True, default=default)

	def format(self, record):
		if self.include_b:
			row = {k: self.myfields[k]() for k in self.include_b}
			row.update(record.msg)
		else:
			row = record.msg

		return self.dumps(row)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewitems

from copy import copy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Any

def cast(object, class_, instanceof=object, *args, **kwargs):
	""" Changes the class of `object` to `class_` if `object` is an instance of `instanceof`,
		calls the initializer and returns it.
	"""

	object = copy(object)
	if isinstance(object, instanceof):
		object.__class__ = class_
		object.__init__(*args, **kwargs)
	else:
		raise TypeError("Object is not an instance of {}".format(instanceof.__name__))
	return object

class STAR(object):
	pass

def args_to_key(args, kwargs, separator=STAR):
	# type: (tuple, dict, Any) -> tuple

	""" Create cache key from function arguments.
	"""

	key = []
	if args:
		key.extend(args)
	if kwargs:
		key.append(separator)
		key.extend(sorted(kwargs.items()))

	return tuple(key)

def compress(value):
	# type: (Any, ) -> Any

	""" Creates a copy of the object where some data structures are replaced with equivalent ones
		which take up less space, but are not necessarily mutable anymore.

		tuple < list
		set == frozenset
		bytes < bytearray
	"""

	# sets are not processed because they cannot contain lists or bytearrays anyway.

	if isinstance(value, (tuple, list)): # tuple *can* contain mutables
		return tuple(compress(x) for x in value)
	elif isinstance(value, bytearray):
		return bytes(value) # bytearray can only be bytes or List[int] right?
	elif isinstance(value, dict):
		return {k: compress(v) for k, v in viewitems(value)}
	else:
		return value



from __future__ import absolute_import, division, print_function, unicode_literals

from time import sleep
from itertools import count
from socket import timeout as SocketTimeout

from hyper import HTTP20Connection
from hyper.http20.exceptions import StreamResetError

from .http import HTTPError

class WrappedHTTP20Connection(object):

	def __init__(self, host):
		self.host = host

	def connect(self):
		self.conn = HTTP20Connection(self.host)

	def _dl_url(self, path, headers, timeout):
		self.conn.request("GET", path, headers=headers)
		self.conn._sock._sck.settimeout(timeout)
		resp = self.conn.get_response()
		if resp.status >= 200 and resp.status < 300:
			return resp.read()

		print(path, resp.status, resp.trailers, resp.headers)
		raise HTTPError("non-2xx error", response=resp)

	def dl_url(self, path, headers=None, timeout=60):
		for i in count(1):
			sleep(i)
			try:
				return self._dl_url(path, headers, timeout)
			except (ConnectionResetError, SocketTimeout, StreamResetError) as e:
				self.connect()
				continue




from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range
from typing import TYPE_CHECKING

from .exceptions import assert_choice

if TYPE_CHECKING:
	from typing import Iterable, Iterator

def encode_binary(boolit, pad="0"):
	# type: (Union[str, Iterator[bool]], str) -> bytes

	""" Encode a string consisting of 0s and 1s or an iterable returns boolean values
		to bytes.

		Example:
		encode_binary("1111111100000001") -> b'\xff\x01'
		encode_binary([True]*8+[False]*7+[True]) -> b'\xff\x01'
	"""

	assert_choice("pad", pad, {"0", "1"})

	if isinstance(boolit, str):
		bin = boolit
	else:
		bin = "".join("1" if b else "0" for b in boolit)

	bin += pad * ((8 - len(bin)%8) % 8) # pad
	assert len(bin) % 8 == 0
	return bytes(int(bin[x:x+8], 2) for x in range(0, len(bin), 8))

def _str2bool_it(s):
	# type: (Iterable[str], ) -> Iterator[bool]

	for chunk in s:
		for c in chunk:
			yield True if c == "1" else False

def decode_binary(key, tostring=False):
	# type: (bytes, bool) -> Union[str, Iterator[bool]]

	""" Decode bytes to either a string of 0s and 1s or an iterable of booleans.
	"""

	it = ("{:08b}".format(b) for b in key)

	if tostring:
		return "".join(it)
	else:
		return _str2bool_it(it)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from datetime import timedelta
from fractions import Fraction
from typing import TYPE_CHECKING

from .compat import FileExistsError
from .compat.os import PathLike, fspath

if TYPE_CHECKING:
	from typing import Any, Union, Iterator, Tuple, Sequence
	import numpy as np
	from av import VideoFrame
	from .compat.pathlib import Path
	WsgiApp = Any

#if __debug__:
#	import av
#	import cv2

logger = logging.getLogger(__name__)

def _raise_exists(outpath, overwrite=False):
	# type: (Path, bool) -> None

	if not overwrite and outpath.exists():
		raise FileExistsError()

class NoGoodFrame(Exception):
	pass

class NoKeyFrame(Exception):
	pass

class BadFile(Exception):
	pass

class VideoBase(object):

	single_frame_ratio = 0.5

	def calculate_offsets(self, time_base, duration):
		# type: (Fraction, int) -> Iterator[int]

		raise NotImplementedError

	def _get_frame(self, offset, native=False):
		# type: (int, bool) -> Tuple[float, np.ndarray]

		raise NotImplementedError

	def _frame_to_file(self, frame, outpath):
		# type: (Any, PathLike) -> None

		raise NotImplementedError

	def iterate(self):
		# type: () -> Iterator[Tuple[float, np.ndarray]]

		for offset in self.calculate_offsets(self.time_base, self.native_duration):
			try:
				yield self._get_frame(offset)
			except NoKeyFrame as e:
				yield offset, e

	def save_frame_to_file(self, pos, outpath):
		# type: (float, PathLike) -> None

		frametime, frame = self._get_frame(int(self.native_duration * pos), native=True)
		self._frame_to_file(frame, outpath)

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.close()

def fourcc(integer):
	import struct
	return struct.pack("<i", integer)

	# return integer.to_bytes(4, "little") # python 3 only

class CvVideo(VideoBase):

	def __init__(self, path):
		# type: (Union[str, PathLike, int], ) -> None

		self.cv2 = self.import_backend()

		if isinstance(path, PathLike):
			path = fspath(path)

		self.cap = self.cv2.VideoCapture(path)

		frame_width = int(self.cap.get(self.cv2.CAP_PROP_FRAME_WIDTH))
		frame_height = int(self.cap.get(self.cv2.CAP_PROP_FRAME_HEIGHT))
		frame_count = int(self.cap.get(self.cv2.CAP_PROP_FRAME_COUNT))

		pixel_aspect_ratio = Fraction(
			int(self.cap.get(self.cv2.CAP_PROP_SAR_NUM)),
			int(self.cap.get(self.cv2.CAP_PROP_SAR_DEN))
		)

		if pixel_aspect_ratio == 0:
			pixel_aspect_ratio = Fraction(1, 1)

		display_aspect_ratio = pixel_aspect_ratio * Fraction(frame_width, frame_height)

		fps = self.cap.get(self.cv2.CAP_PROP_FPS)
		try:
			duration = timedelta(seconds=frame_count / fps)
		except ZeroDivisionError:
			raise BadFile("Cannot open {}".format(path))

		self.native_duration = frame_count # exclusive
		self.time_base = Fraction(1, 1)
		self.meta = {
			"width": frame_width,
			"height": frame_height,
			"duration": duration, # duration in seconds
			"fps": fps,
			"sample_aspect_ratio": pixel_aspect_ratio,
			"display_aspect_ratio": display_aspect_ratio,
			"format": fourcc(int(self.cap.get(self.cv2.CAP_PROP_CODEC_PIXEL_FORMAT))).decode("ascii"),
		}

	@staticmethod
	def import_backend():
		import cv2

		return cv2

	def time_to_seconds(self, offset):
		# type: (int, ) -> float

		return offset / self.meta["fps"]

	def _get_frame(self, offset, native=False):
		# type: (int, ) -> Tuple[float, np.ndarray]

		self.cap.set(self.cv2.CAP_PROP_POS_FRAMES, offset)

		ret, frame = self.cap.read()
		if ret:
			if not native:
				frame = self.cv2.cvtColor(frame, self.cv2.COLOR_BGR2RGB)
			return self.time_to_seconds(offset), frame
		else:
			raise NoGoodFrame("Could not find good frame")

	def _frame_to_file(self, frame, outpath):
		# type: (np.ndarray, PathLike) -> None

		is_success, im_buf = self.cv2.imencode(outpath.suffix, frame) # type: Tuple[bool, np.ndarray]
		assert is_success
		im_buf.tofile(fspath(outpath)) # cv2.imwrite doesn't handle unicode paths correctly

	def close(self):
		# type: () -> None

		self.cap.release()

def object_attributes(obj):

	def asd():
		for attr in dir(obj):
			val = getattr(obj, attr)
			if not callable(val):
				yield attr, val

	return dict(asd())

class AvVideo(VideoBase):

	vcc_attrs = ["coded_height", "coded_width", "display_aspect_ratio", "encoded_frame_count", "format",
			"framerate", "gop_size", "has_b_frames", "height", "pix_fmt", "reformatter",
			"sample_aspect_ratio", "width"]

	def __init__(self, path, videostream=0):
		# type: (PathLike, int) -> None

		self.av = self.import_backend()

		if isinstance(path, PathLike):
			path = fspath(path)

		try:
			self.container = self.av.open(path, "r")
		except self.av.error.InvalidDataError:
			raise BadFile("Cannot open {}".format(path))

		if self.container.format.name == "matroska,webm":
			raise BadFile("Matroska files are currently not supported :(")

		self.vstream = self.container.streams.video[videostream]
		self.vstream.thread_type = "AUTO"

		duration = timedelta(seconds=self.container.duration / self.av.time_base)

		vcc = self.vstream.codec_context
		vcc = {attr:getattr(vcc, attr) for attr in self.vcc_attrs}
		del vcc["format"]

		self.native_duration = self.vstream.duration # exclusive
		self.time_base = self.vstream.time_base

		if not self.native_duration:
			self.native_duration = self.container.duration
			self.time_base = Fraction(1, self.av.time_base)
			logger.debug("Using container instead of video stream duration")

		if not self.native_duration:
			raise BadFile("Cannot open {}".format(path))

		# what to do with self.vstream.start_time?

		self.meta = {
			"width": vcc["width"],
			"height": vcc["height"],
			"duration": duration,
			"fps": vcc["framerate"],
			"sample_aspect_ratio": vcc["sample_aspect_ratio"] or Fraction(1, 1),
			"display_aspect_ratio": vcc["display_aspect_ratio"] or Fraction(vcc["width"], vcc["height"]),
			"format": vcc["pix_fmt"],
			#"containser_bit_rate": self.container.bit_rate,
		}

	@staticmethod
	def import_backend():
		import av

		return av

	def _get_frame(self, offset, native=False):
		# type: (int, bool) -> Tuple[float, np.ndarray]

		for i in range(1): # trying x times to find good frames
			try:
				self.container.seek(offset, backward=True, any_frame=False, stream=self.vstream) # this can silently fail for broken files
			except self.av.error.PermissionError:
				raise NoKeyFrame("Failed to seek to {} of {}.".format(offset, self.container.duration))

			try:
				for vframe in self.container.decode(self.vstream): # can raise
					if not vframe.is_corrupt:
						logger.debug("Grabbing frame at %.03fs from seek offset %.03fs", vframe.time, offset*self.time_base)

						if native:
							frame = vframe
						else:
							frame = vframe.to_ndarray(format="rgb24")

						return vframe.time, frame

					logger.debug("Skipping corrupt frame at %.03fs from seek offset %.03fs", vframe.time, offset*self.time_base)

			except (self.av.error.InvalidDataError, self.av.error.ValueError) as e:
				logger.warning("Skipping corrupt container position at seek offset %.03fs: %s", offset*self.time_base, e)
				offset += int(self.time_base.denominator / self.meta["fps"])
			else:
				raise NoGoodFrame("Could not find good frame after decoding full file")

		raise NoGoodFrame("Could not find good frame after trying various offsets")

	def _frame_to_file(self, frame, outpath):
		# type: (VideoFrame, Path) -> None

		frame.to_image(fspath(outpath))

	def close(self):
		# type: () -> None

		self.container.close()

def grab_pic(inpath, outpath, pos=0.5, overwrite=False, backend="cv"):
	# type: (Union[str, PathLike], Path, Union[float, Sequence[float]], bool, str) -> None

	if backend == "av":
		vf = AvVideo(inpath)
	elif backend == "cv":
		vf = CvVideo(inpath)
	else:
		raise ValueError("Unsupported backend: {}".format(backend))

	try:
		outpath.parent.mkdir(parents=True, exist_ok=True)

		if isinstance(pos, float):
			_raise_exists(outpath, overwrite)
			vf.save_frame_to_file(pos, outpath)
		elif isinstance(pos, list):
			for i, p in enumerate(pos):
				outpathseq = outpath.with_suffix(".{}{}".format(i, outpath.suffix))
				_raise_exists(outpathseq, overwrite)
				vf.save_frame_to_file(p, outpathseq)
		else:
			raise TypeError("pos")

	finally:
		vf.close()

if __name__ == "__main__":

	from argparse import ArgumentParser
	parser = ArgumentParser()
	parser.add_argument("path")
	args = parser.parse_args()

	with CvVideo(args.path) as vf:
		print(vf.meta)

	with AvVideo(args.path) as vf:
		print(vf.meta)



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import range

from random import choice, randrange
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Iterator, Tuple

def randstr(length, charset):
	# type: (int, str) -> str

	""" Returns a (noncryptographic) random string consisting of characters from `charset`
		of length `length`.
	"""

	return "".join(choice(charset) for i in range(length))  # nosec

def randbytes(size):
	# type: (int, ) -> bytes

	""" Returns (noncryptographic) random bytes of length `length`.
	"""

	return b"".join(chr(randrange(0, 256)) for _ in range(size))  # nosec

def rgb_colors():
	# type: () -> Iterator[Tuple[int, int, int]]

	""" Yields a stream of (noncryptographic) random RGB color tuples.
	"""

	while True:
		rgb = randrange(0, 256**3)  # nosec
		rg, b = divmod(rgb, 256)
		r, g = divmod(rg, 256)
		yield (r, g, b)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.moves.urllib.parse import urlsplit, urlunsplit, parse_qs, urlencode

import re
from string import ascii_letters, digits
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Optional, Dict, Iterable
	from future.moves.urllib.parse import SplitResult

# URI RFC
gen_delims = ":/?#[]@"
sub_delims = "!$&'()*+;="
reserved = gen_delims + sub_delims
unreserved = ascii_letters + digits + "-._~"
valid_uri_characters = reserved + unreserved

uri_schemes = ("http", "https", "ftp", "sftp", "irc", "magnet", "file", "data")
url_base_pattern = r"(?:(?:{}:\/\/)|www\.)(?:[{}]|%[a-zA-Z0-9]{{2}})+"

def get_url_pattern(schemes=None):
	# type: (Optional[Iterable[str]], ) -> re.Pattern

	schemes = schemes or uri_schemes
	return re.compile(url_base_pattern.format("|".join(schemes), re.escape(valid_uri_characters)))

def get_filename_from_url(url, strip=None):
	# type: (str, Optional[str]) -> str

	return urlsplit(url).path.rstrip(strip).rsplit("/", 1)[1]

def get_url_argument(split_url, argument, path=None):
	# type: (SplitResult, str, Optional[str]) -> Optional[str]

	""" Extracts a parameter value from the url query string. Optionally compares the url path.
		For example:
		> get_url_argument(urlsplit("https://www.google.com/search?q=python"), "q")
		'python'
	"""

	if not path or split_url.path == path:
		try:
			return parse_qs(split_url.query)[argument][0]
		except KeyError:
			return None
	return None

def url_replace_query(split_url, query, drop_fragment=True):
	# type: (SplitResult, Dict[str, str], bool) -> str

	""" Replace the query part of an URL with a new one.
		For example:
		> url_replace_query(urlsplit("https://www.google.com/search?source=hp&q=java"), {"q": "python"})
		'https://www.google.com/search?q=python'
	"""

	scheme, netloc, path, _, fragment = split_url

	query = urlencode(query)
	if drop_fragment:
		fragment = ""

	return urlunsplit((scheme, netloc, path, query, fragment))



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from OpenSSL import crypto  # pyOpenSSL
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat, PublicFormat, NoEncryption
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.serialization import load_pem_private_key, load_der_private_key
from cryptography.hazmat.backends import default_backend

from .exceptions import assert_choice

if TYPE_CHECKING:
	from typing import Union
	from OpenSSL.crypto import X509

	from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
	from cryptography.hazmat.primitives.asymmetric.dsa import DSAPrivateKey
	from cryptography.hazmat.primitives.asymmetric.dh import DHPrivateKey
	from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
	PrivateKey = Union[RSAPrivateKey, DSAPrivateKey, DHPrivateKey, EllipticCurvePrivateKey]

def load_certificate_file(filename):
	# type: (str, ) -> X509

	with open(filename, "rb") as fr:
		return crypto.load_certificate(crypto.FILETYPE_PEM, fr.read())

def get_pubkey_from_x509(x509):
	# type: (X509, ) -> bytes

	return crypto.dump_privatekey(crypto.FILETYPE_ASN1, x509.get_pubkey())

def load_keyfile(filename, encoding="PEM"):
	# type: (str, str) -> PrivateKey

	try:
		load_func = {
			"PEM": load_pem_private_key,
			"DER": load_der_private_key,
		}[encoding]
	except KeyError:
		raise ValueError("Invalid encoding")

	with open(filename, "rb") as fr:
		password = None
		backend = default_backend()
		return load_func(fr.read(), password, backend)

def generate_rsa_keyfile_pair(priv_key, pub_key, key_size=4096, encoding="PEM", format="modern"):
	# type: (str, str, int, str, str) -> RSAPrivateKey

	"""
		encoding: "PEM" -> ASN.1 encoding, "DER" -> base64 PEM
	"""

	if key_size < 2048:
		raise ValueError("key_size must be atleast 2048")

	encodings = {
		"PEM": Encoding.PEM,
		"DER": Encoding.DER,
	}

	private_formats = {
		"modern": PrivateFormat.PKCS8,
		"legacy": PrivateFormat.TraditionalOpenSSL,
	}

	public_formats = {
		"modern": PublicFormat.SubjectPublicKeyInfo,
		"legacy": PublicFormat.PKCS1,
	}

	assert_choice("encoding", encoding, encodings)
	assert_choice("format", format, private_formats)

	public_exponent = 65537
	backend = default_backend()

	key = generate_private_key(public_exponent, key_size, backend)

	with open(priv_key, "wb") as fw:
		fw.write(key.private_bytes(encodings[encoding], private_formats[format], NoEncryption()))

	with open(pub_key, "wb") as fw:
		fw.write(key.public_key().public_bytes(encodings[encoding], public_formats[format]))

	return key



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewitems
from collections import defaultdict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Any, Callable, Dict, Hashable, Iterable, Iterator, Mapping, Optional, Tuple, TypeVar

	T = TypeVar("T")
	U = TypeVar("U")
	H = TypeVar("H", bound=Hashable)
	V = TypeVar("V")

def flatten(d):
	# type: (Dict[T, U], ) -> Iterator[U]

	""" Flattens dicts of (dicts of...) lists to lists. """

	if isinstance(d, (list, tuple)):
		for v in d:
			yield v
	elif isinstance(d, dict):
		for k, val in viewitems(d):
			for v in flatten(val):
				yield v
	else:
		raise TypeError("Unsupported type: {}".format(type(d)))

def get_one_of(d, keys):
	# type: (Dict[T, U], Iterable[T]) -> Tuple[T, U]

	""" Returns the (key, value) pair of the first key of `keys` found in `d`.
	"""

	for key in keys:
		try:
			return key, d[key]
		except KeyError:
			pass

	raise KeyError("None of the {} keys could be found".format(len(keys)))

def get_available(d, keys):
	# type: (Dict[T, U], Iterable[T]) -> Iterator[Tuple[T, U]]

	for key in keys:
		try:
			yield key, d[key]
		except KeyError:
			pass

def hasvalues(d):
	# type: (dict, ) -> dict

	return {k:v for k, v in viewitems(d) if v}

def valuemap(func, d):
	# type: (Callable[[U], V], Dict[T, U]) -> Dict[T, V]

	return {k: func(v) for k, v in viewitems(d)}

# was: mapdict, mapget
def mapmap(d, it):
	# type: (Mapping[T, U], Iterable[T]) -> Iterator[U]

	return (d[i] for i in it)

def itemgetter(it):
	# type: (Iterable[T], ) -> Callable[[Mapping[T, U]], Iterator[U]]

	""" Similar to `operator.itemgetter` except that it always expects and returns iterables.
		Compare `mapmap`
	"""

	return lambda d: (d[i] for i in it)

def subdict(d, it):
	# type: (Mapping[T, U], Iterable[T]) -> Dict[T, U]

	""" Uses the elements of `it` as keys to extract a new sub-dictionary. """

	return {key: d[key] for key in it}

def subdictdefault(d, it, default=None):
	# type: (Mapping[T, U], Iterable[T], Optional[T]) -> Dict[T, U]

	""" Uses the elements of `it` as keys to extract a new sub-dictionary. """

	return {key: d.get(key, default) for key in it}

def update(d1, d2):
	# type: (dict, dict) -> None

	""" Same as `dict.update` except that `None` values are skipped. """

	for k, v in viewitems(d2):
		if v is not None:
			d1[k] = v

class keydefaultdict(defaultdict):

	""" defaultdict which passes a key to the default factory. """

	def __missing__(self, key):
		# type: (Hashable, ) -> Any

		if self.default_factory is None:
			raise KeyError(key)
		else:
			value = self[key] = self.default_factory(key)
			return value



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map, chr, range

from unicodedata import category
from sys import maxunicode
from collections import defaultdict
from tempfile import gettempdir

from .pickle import cache  # nosec

@cache(gettempdir() + "/unicode_categories.{ppv}.pkl", ignoreargs=True)
def unicode_categories():
	unicode_categories = defaultdict(set)
	for c in map(chr, range(maxunicode + 1)):
		unicode_categories[category(c)].add(c)
	return unicode_categories



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewkeys
import json
from typing import TYPE_CHECKING

from .http import URLRequest
from .ops import logical_xor

if TYPE_CHECKING:
	from typing import Callable, Optional, List, Tuple

class StreamWatcher(object):

	def __init__(self, api):
		self.api = api
		self.followed_names = api.get_followed()
		self.followed_online = {userid: False for userid in viewkeys(self.followed_names)}

	def watch(self, notify_started, notify_stopped):
		# type: (Callable[[str, str, Optional[str]], None], Callable[[str, str], None]) -> None

		user_ids = viewkeys(self.followed_names)
		streams = self.api.get_streams(user_ids)
		online = viewkeys(streams)
		offline = user_ids - online

		for user_id in online:
			if not self.followed_online[user_id]:
				self.followed_online[user_id] = True
				notify_started(user_id, self.followed_names[user_id], streams[user_id].get("title", None))

		for user_id in offline:
			if self.followed_online[user_id]:
				self.followed_online[user_id] = False
				notify_stopped(user_id, self.followed_names[user_id])

class TwitchAPI(object):

	base = "https://api.twitch.tv/helix/"
	login = "users"
	follows = "users/follows"
	streams = "streams?user_id={}"

	def __init__(self, client_id, userid=None, username=None):
		# type: (str, Optional[str], Optional[str]) -> None

		if not logical_xor(userid, username):
			raise ValueError("Either the userid or the username must be given")

		self.client_id = client_id

		if userid:
			self.userid = userid
		else:
			self.userid = self.get_userid(username)

	def req(self, url, params):
		# type: (str, List[Tuple[str, str]]) -> dict

		qs = "&".join(k+"="+v for k, v in params)
		data = URLRequest(url + "?" + qs, headers={"Client-ID": self.client_id}).load()
		return json.loads(data)

	def get_userid(self, username):
		# type: (str, ) -> str

		d = self.req(self.base + self.login, [("login", username)])
		return d["data"][0]["id"]

	def get_followed(self):
		d = self.req(self.base + self.follows, [("from_id", self.userid)])
		return {follow["to_id"]: follow["to_name"] for follow in d["data"]}

	def get_streams(self, user_ids):
		d = self.req(self.base + self.streams, [("user_id", user_id) for user_id in user_ids])
		ret = {stream["user_id"]: stream for stream in d["data"]}
		assert len(d["data"]) == len(ret), "More than one stream per user"
		return ret

	def watcher(self):
		# type: () -> StreamWatcher

		return StreamWatcher(self)



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from .exceptions import NoResult, InconsistentState

if TYPE_CHECKING:
	from pymongo.collection import Collection

def findone(db, query):
	# type: (Collection, str) -> dict

	docs = list(db.find(query))

	if len(docs) == 0:
		raise NoResult("No result found")
	elif len(docs) == 1:
		return docs[0]
	else:
		raise InconsistentState("More than one result found")



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from pymediainfo import MediaInfo

from .compat.os import fspath, PathLike

if TYPE_CHECKING:
	from typing import Union
	from pymediainfo import Track
	from .compat.pathlib import Path

class MediaInfoHelper(object):

	def __init__(self, path):
		# type: (Union[str, Path], ) -> None

		if isinstance(path, PathLike):
			path = fspath(path)

		self.mi = MediaInfo.parse(path)

	def audio_duration(self):
		# type: () -> float

		""" Returns audio duration in seconds. """

		for track in self.mi.tracks:
			if track.track_type == "Audio":
				return track.duration / 1000.

		raise LookupError("Audio duration not available")

	@property
	def general(self):
		return self.first_track("General")

	@property
	def video(self):
		return self.first_track("Video")

	@property
	def audio(self):
		return self.first_track("Audio")

	@property
	def text(self):
		return self.first_track("Text")

	def first_track(self, track_type):
		# type: (str, ) -> Track

		for track in self.mi.tracks:
			if track.track_type == track_type:
				return track

		raise LookupError("Video track not available")

	def is_vfr(self):
		for track in self.mi.tracks:
			if track.track_type == "Video":
				return track.frame_rate_mode == "VFR"

		return False

	def meta_info(self):

		for track in self.mi.tracks:
			if track.track_type == "General":

				if track.duration is None:
					duration = None
				else:
					duration = track.duration / 1000.

				return {
					"title": track.title, # track_name
					"performer": track.performer,
					"album": track.album,
					"track-position": track.track_name_position,
					"duration": duration,
					"date": track.Recorded_Date,
				}

		raise LookupError("Meta data not available")



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Union

try:
	from os import makedirs

except ImportError:
	import os, errno

	def makedirs(name, mode=0o777, exist_ok=False):
		try:
			return os.makedirs(str(name), mode)
		except OSError as e:
			if exist_ok and e.errno == errno.EEXIST:
				pass
			else:
				raise

try:
	from os import PathLike

except ImportError:
	from abc import abstractmethod
	from .abc import ABC

	class PathLike(ABC):

		"""Abstract base class for implementing the file system path protocol."""

		@abstractmethod
		def __fspath__(self):
			"""Return the file system path representation of the object."""
			raise NotImplementedError

		@classmethod
		def __subclasshook__(cls, subclass):
			return hasattr(subclass, "__fspath__")

try:
	from os import replace

except ImportError:

	import sys

	if sys.platform == "win32":
		from ctypes import WinError
		from cwinsdk.um.WinBase import MOVEFILE_REPLACE_EXISTING

		def replace(src, dst):
			# type: (Union[str, bytes], Union[str, bytes]) -> None

			if isinstance(src, str) and isinstance(dst, str):
				from cwinsdk.um.WinBase import MoveFileExW as MoveFileEx
			elif isinstance(src, bytes) and isinstance(dst, bytes):
				from cwinsdk.um.WinBase import MoveFileExA as MoveFileEx
			else:
				raise ValueError("Arguments must be both bytes or both string")

			if MoveFileEx(src, dst, MOVEFILE_REPLACE_EXISTING) == 0:
				raise WinError()

	else:
		from os import rename as replace # on posix rename can already replace existing files atomically

try:
	from os import scandir as _scandir
except ImportError:
	from scandir import scandir as _scandir

if hasattr(_scandir, "__exit__"):
	scandir = _scandir
else:
	class scandir(object):

		def __init__(self, path="."):
			self.it = _scandir(path)

		def __enter__(self):
			return self.it

		def __exit__(self, *args):
			pass

		def __iter__(self):
			return self.it

		def __next__(self):
			return next(self.it)

		next = __next__

try:
	from os import DirEntry

except ImportError:
	DirEntry = type(next(scandir()))

try:
	from os import fspath

except ImportError:

	fspath = str # bad, but at least works for `Path`s.



import logging, re, csv
from time import sleep
from collections.abc import Sequence
from typing import TYPE_CHECKING

import requests
from simple_salesforce import Salesforce
from simple_salesforce.exceptions import SalesforceExpiredSession, SalesforceAuthenticationFailed
from simplejson.errors import JSONDecodeError

from .iter import progress
from .json import read_json, write_json

if TYPE_CHECKING:
	from typing import List, Iterable, Optional, Tuple, Iterator

logger = logging.getLogger(__name__)

_sosl_pat = re.compile("[" + re.escape("?&|!{}[]()^~*:\\\"'+-") + "]")
_sosl_repl = lambda m: "\\" + m.group(0)

def sosl_escape(s):
	# type: (str, ) -> str

	return _sosl_pat.sub(_sosl_repl, s)

class MySalesforce(object):

	timeout = 60

	def __init__(self, username, password, security_token, consumer_key, consumer_secret, test=False, cache_file=None, timeout=60):
		# type: (str, str, str, str, str, bool, Optional[str], int) -> None

		self.username = username
		self.password = password
		self.security_token = security_token
		self.consumer_key = consumer_key
		self.consumer_secret = consumer_secret
		self.test = test
		self.cache_file = cache_file
		self.timeout = timeout

		self._session = None # type: Optional[Salesforce]

	def _login(self):
		# type: () -> Tuple[str, str]

		logger.debug("Obtaining new Salesforce credentials")

		if self.test:
			url = "https://test.salesforce.com/services/oauth2/token"
		else:
			url = "https://login.salesforce.com/services/oauth2/token"

		params = {
			"grant_type": "password",
			"client_id": self.consumer_key,
			"client_secret": self.consumer_secret,
			"username": self.username,
			"password": self.password + self.security_token,
		}

		r = requests.post(url, params=params, timeout=self.timeout)
		r.raise_for_status()
		response = r.json()

		return response["instance_url"], response["access_token"]

	def session(self, fresh=False):
		# type: () -> Salesforce

		if self._session is not None and not fresh:
			return self._session

		try:
			if fresh or not self.cache_file:
				raise FileNotFoundError

			obj = read_json(self.cache_file)
			logger.debug("Creating new Salesforce session")
			self._session = Salesforce(instance_url=obj["instance_url"], session_id=obj["session_id"])

		except FileNotFoundError:
			instance_url, session_id = self._login()

			if self.cache_file:
				write_json({
					"instance_url": instance_url,
					"session_id": session_id,
				}, self.cache_file)

			logger.debug("Creating new Salesforce session")
			self._session = Salesforce(instance_url=instance_url, session_id=session_id)

		return self._session

	# internal query/search

	def _query(self, s):
		# type: (str, ) -> dict

		try:
			return self.session().query(s)
		except SalesforceExpiredSession:
			logger.debug("Salesforce session expired")
			return self.session(True).query(s)

	def _query_all(self, s):
		# type: (str, ) -> Iterator[dict]

		reconnect = True

		try:
			for row in self.session().query_all_iter(s):
				yield row
				reconnect = False
		except SalesforceExpiredSession:
			logger.debug("Salesforce session expired")
			if reconnect:
				for row in self.session(True).query_all_iter(s):
					yield row
			else:
				raise RuntimeError("Cannot reconnect Salesforce session after partial query")

	def _search(self, s):

		try:
			return self.session().search(s)
		except SalesforceExpiredSession:
			logger.debug("Salesforce session expired")
			return self.session(True).search(s)

	def rest_post(self, endpoint, params):
		# type: (str, dict) -> dict

		sess = self.session()
		headers = {"Authorization": "Bearer " + sess.session_id}

		r = requests.post("https://" + sess.sf_instance + endpoint, headers=headers, json=params)
		r.raise_for_status()
		return r.json()

	def rest_get(self, endpoint):
		# type: (str, ) -> dict

		sess = self.session()
		headers = {"Authorization": "Bearer " + sess.session_id}

		r = requests.get("https://" + sess.sf_instance + endpoint, headers=headers)
		r.raise_for_status()
		return r.json()

	# actual methods

	def query(self, query_str):
		# type: (str, ) -> dict

		return self._query(query_str)

	def get_all_fields(self, object_name):
		# type: (str, ) -> dict

		""" Retrieves all fields of `object_name`.
			Warning: `object_name` is not escaped!
		"""

		query_str = "SELECT QualifiedApiName FROM FieldDefinition WHERE EntityDefinition.QualifiedApiName = '{}'".format(object_name)  # nosec

		return self._query(query_str)

	def get_all_objects(self):

		query_str = "SELECT QualifiedApiName FROM EntityDefinition ORDER BY QualifiedApiName"

		return self._query(query_str)

	def search_fields(self, s, object_name, object_fields):
		# type: (str, str, Iterable[str]) -> dict

		""" Searches for `s` in object_name with object_fields.
			Warning: `object_name` and `object_fields` are not escaped!
		"""

		s = sosl_escape(s)
		return self._search("FIND {{{}}} RETURNING {}({})".format(s, object_name, ", ".join(object_fields)))

	def dump_csv(self, query_str, path, verbose=False):
		# type: (str, str, False) -> int

		""" Run SOQL `query_str` and dump results to csv file `path`.
			Returns the number of exported rows.
		"""

		i = 0

		with open(path, "w", encoding="utf-8", newline="") as csvfile:
			csvwriter = csv.writer(csvfile)

			if verbose:
				it = progress(self._query_all(query_str))
			else:
				it = self._query_all(query_str)

			row = next(it)
			del row["attributes"]
			csvwriter.writerow(row.keys())
			csvwriter.writerow(row.values())
			i += 1

			for row in it:
				del row["attributes"]
				csvwriter.writerow(row.values())
				i += 1

		return i

class LiveAgent(object):

	# https://help.salesforce.com/articleView?id=000331168&type=1&mode=1
	# https://help.salesforce.com/articleView?id=000340657&type=1&mode=1

	api_version = "42"

	def __init__(self, hostname, organization_id, deployment_id, button_id, timeout=60):
		# type: (str, str, str, str, int) -> None

		self.hostname = hostname
		self.organization_id = organization_id
		self.deployment_id = deployment_id
		self.button_id = button_id
		self.timeout = timeout

		self.key = None
		self.affinity_token = None
		self._sequence = 0

	# helper

	def urljoin(self, endpoint):
		return "https://" + self.hostname + endpoint

	@property
	def sequence(self):
		self._sequence += 1
		return str(self._sequence)

	# high level

	def connect(self, visitor_name, slots=None):
		# type: (str, Optional[Union[Dict[str, str], Sequence[Dict[str, Any]]]]) -> None

		""" Connect using `visitor_name` as name.
		"""

		response = self.sessionid()
		slots = slots or {}

		self.key = response["key"]
		self.affinity_token = response["affinityToken"]
		session_id = response["id"]

		if isinstance(slots, dict):
			prechat_details = []  # type: List[Dict[str, Any]]

			for label, value in slots.items():
				custom_detail = {
					"label": label,
					"value": value,
					"displayToAgent": True,
					"transcriptFields": [],
					"entityFieldMaps": [],
				}
				prechat_details.append(custom_detail)

		elif isinstance(slots, Sequence):
			try:
				for custom_detail in slots:
					custom_detail["label"]
					custom_detail["value"]
					custom_detail["displayToAgent"]
			except KeyError:
				raise ValueError("slots is missing fields")

			prechat_details = slots

		else:
			raise ValueError("slots must be dict or Sequence")

		self.chasitorinit(self.key, self.affinity_token, session_id, visitor_name, prechat_details=prechat_details)

	def is_available(self):
		# type: () -> bool

		""" Check for agent availability.
		"""

		res = self.availability()
		for msg in res["messages"]:
			if msg["type"] == "Availability":
				return msg["message"]["results"][0].get("isAvailable", False)
			else:
				logger.error("Got unexpected message: %s", msg)

	def wait_available(self, wait=10):
		# type: (float, ) -> None

		while True:
			if self.is_available():
				return

			sleep(wait)

	def receive(self):
		# type: () -> List[dict]

		""" Long poll messages.
		"""

		while True:
			r = self.messages(self.key, self.affinity_token)
			try:
				return r.json()["messages"]
			except JSONDecodeError:
				pass

	def send(self, text):
		# type: (str, ) -> None

		""" Send text message to agent.
		"""

		self.chatmessage(self.key, self.affinity_token, text)

	def close(self):
		# type: () -> dict

		""" Close live chat.
		"""

		return self.chatend(self.key, self.affinity_token)

	# low level

	def sessionid(self):
		# type: () -> dict

		endpoint = "/chat/rest/System/SessionId"

		headers = {
			"X-LIVEAGENT-API-VERSION": self.api_version,
			"X-LIVEAGENT-AFFINITY": "null",
		}

		r = requests.get(self.urljoin(endpoint), headers=headers, timeout=self.timeout)
		r.raise_for_status()
		return r.json()

	def chasitorinit(self, key, affinity_token, session_id, visitor_name,
		user_agent="", language="en-US", screen_resolution="1920x1080", prechat_details=()):
		# type: (str, str, str, str, str, str, str, Sequence[Dict[str, Any]]) -> requests.Request

		endpoint = "/chat/rest/Chasitor/ChasitorInit"

		headers = {
			"X-LIVEAGENT-API-VERSION": self.api_version,
			"X-LIVEAGENT-AFFINITY": affinity_token,
			"X-LIVEAGENT-SESSION-KEY": key,
			"X-LIVEAGENT-SEQUENCE": self.sequence,
		}

		params = {
			"organizationId": self.organization_id,
			"deploymentId": self.deployment_id,
			"buttonId": self.button_id,
			"sessionId": session_id,
			"userAgent": user_agent,
			"language": language,
			"screenResolution": screen_resolution,
			"visitorName": visitor_name,
			"prechatDetails": prechat_details,
			"prechatEntities": [],
			"receiveQueueUpdates": True,
			"isPost": True,
		}
		print(params)
		r = requests.post(self.urljoin(endpoint), headers=headers, json=params, timeout=self.timeout)
		r.raise_for_status()
		return r

	def messages(self, key, affinity_token):

		endpoint = "/chat/rest/System/Messages"

		headers = {
			"X-LIVEAGENT-API-VERSION": self.api_version,
			"X-LIVEAGENT-AFFINITY": affinity_token,
			"X-LIVEAGENT-SESSION-KEY": key,
		}

		r = requests.get(self.urljoin(endpoint), headers=headers, timeout=self.timeout)
		r.raise_for_status()
		return r

	def chatmessage(self, key, affinity_token, text):

		endpoint = "/chat/rest/Chasitor/ChatMessage"

		headers = {
			"X-LIVEAGENT-API-VERSION": self.api_version,
			"X-LIVEAGENT-AFFINITY": affinity_token,
			"X-LIVEAGENT-SESSION-KEY": key,
			"X-LIVEAGENT-SEQUENCE": self.sequence,
		}

		params = {
			"text" : text,
		}

		r = requests.post(self.urljoin(endpoint), headers=headers, json=params, timeout=self.timeout)
		r.raise_for_status()
		return r

	def chatend(self, key, affinity_token):
		# type: (str, str) -> dict

		endpoint = "/chat/rest/Chasitor/ChatEnd"

		headers = {
			"X-LIVEAGENT-API-VERSION": self.api_version,
			"X-LIVEAGENT-AFFINITY": affinity_token,
			"X-LIVEAGENT-SESSION-KEY": key,
			"X-LIVEAGENT-SEQUENCE": self.sequence,
		}

		params = {
			"reason": "client",
		}

		r = requests.post(self.urljoin(endpoint), headers=headers, json=params, timeout=self.timeout)
		r.raise_for_status()
		try:
			return r.json()  # fixme: does this always throw?
		except JSONDecodeError:
			return {}

	def availability(self, estimated_wait_time=False):
		# type: (bool, ) -> dict

		endpoint = "/chat/rest/Visitor/Availability"

		headers = {
			"X-LIVEAGENT-API-VERSION": self.api_version,
		}

		params = {
			"org_id": self.organization_id,
			"deployment_id": self.deployment_id,
			"Availability.ids": [self.button_id],
			"Availability.needEstimatedWaitTime": str(int(estimated_wait_time)),
		}

		r = requests.get(self.urljoin(endpoint), headers=headers, params=params, timeout=self.timeout)
		r.raise_for_status()
		return r.json()



from __future__ import absolute_import, division, print_function, unicode_literals

class ParseError(Exception):
	""" Raised when the content to be parsed is malformed.
		Not a value error because usually the error is in some external resource
		which we don't have any control over.
	"""

class DownloadFailed(Exception):
	""" Raised when a download from an external resource by HTTP or any other protocol did not
		complete successfully. Could be raised anywhere between the initial TCP connection to
		a final integrity check.
	"""

class AccessDenied(Exception):
	pass

class TemporaryError(Exception):
	""" Signals that an failed operation can be retried in the future
		with a possibly different outcome.
	"""

class IteratorExhausted(Exception):
	""" Tried to get data from an exhausted Iterator. """

class WouldBlockForever(Exception):
	""" Raised when a blocking operation would wait forever. """

# results, control flow

class ControlFlowException(Exception):
	""" Not really an error, rather used for control flow handling
		where other constructs like `return` or `break` are not convenient.
	"""

class Skip(ControlFlowException):
	pass

class NoResult(ControlFlowException):
	""" Raised when an operation which maybe returns a result yields no result. """

class Break(ControlFlowException):
	""" Control flow exception to break out of a recursive call. """

class NoActionNeeded(ControlFlowException):
	""" Raised when nothing needs to be done / nothing was modified. """

# external errors

class ExternalError(Exception):
	""" Raised when an error occurs due to an external resource which we don't have control over,
		or are not the only actor to control it.
	"""

# aka ConsistencyError
class InconsistentState(ExternalError):
	""" This is either raised when some external resource changed without our knowledge,
		or when our code violates some consistency assumptions.
	"""

class ExternalProcedureUnavailable(ExternalError):
	""" If some external resource cannot be reached by API or RPC or is otherwise not able to handle
		the request, this is raised. It is not raised for invalid requests to the resource which are
		correctly handled (or rather correctly not handled in thie case).
	"""

# runtime, possible coding errors

class ClosedObjectUsed(RuntimeError):

	def __init__(self, obj):
		RuntimeError.__init__(self, "{} is already closed".format(obj.__class__.__name__))

# values, input errors

class EmptyIterable(ValueError):
	""" Raised when Iterable is passed which doesn't yield any values,
		and thus not resulted can be computed.
	"""

class MalformedFile(ValueError):
	""" Raised when a malformed File is passes as input.
		In contrast to `ParseError` this is usually a file we have control over.
	"""

def assert_choice(name, value, choices):
	# type: (str, T, Set[T]) -> None

	if value not in choices:
		raise ValueError("{} must be one of {}".format(name, ", ".join(choices)))

def assert_type(name, value, types):
	# type: (str, Any, Sequence[Any]) -> None

	if not isinstance(value, types):
		raise TypeError("{} must be one of these types: {}".format(name, ", ".join(types)))

def assert_true(name, value):
	if not value:
		raise ValueError("{} must be set".format(name))



from __future__ import absolute_import, division, print_function, unicode_literals

# needs python 3.4

import sys, asyncio
from time import time
from typing import TYPE_CHECKING

from .iter import progressdata, _lstr

if TYPE_CHECKING:
	from typing import Union, Iterable, Sequence, Optional, IO
	from numbers import Number

class progress_content(object):

	def __init__(self, it, length=None, refresh=1, file=sys.stdout):
		# type: (Union[Iterable, Sequence], Optional[int], Number, Optional[IO[str]]) -> None

		self.it = it
		self.length = length
		self.refresh = refresh
		self.file = file

	def __iter__(self):
		return progressdata(self.it, self.length, self.refresh, file=self.file)

	def __aiter__(self):
		return self.AsyncIterProgress(self.it, self.length, self.refresh, file=self.file)

	class AsyncIterProgress(object):

		def __init__(self, it, length, refresh, file):
			self.it = it.__aiter__()
			self.refresh = refresh
			self.file = file

			self.length, self.lstr = _lstr(self.it, length)
			self.last = self.start = time()
			self.total = 0

		@asyncio.coroutine
		def __anext__(self):
			try:
				elm = yield from self.it.__anext__()
				self.total += len(elm)
				current = time()
				if current - self.last > self.refresh:
					self.last = current
					duration = current - self.start
					print("{}{}, running for {} seconds ({:0.2e}/s).".format(self.total, self.lstr, int(duration), self.total/duration), end="\r", file=self.file)
				return elm

			except StopAsyncIteration:
				print("Finished {} in {} seconds.".format(self.total, int(self.last - self.start)), end="\r", file=self.file)
				raise StopAsyncIteration

if __name__ == "__main__":
	# needs python 3.5

	class gensync:

		def __init__(self):
			self.i = 3

		def __iter__(self):
			return self

		def __next__(self):
			if self.i > 0:
				self.i -= 1
				return "asd"
			else:
				raise StopIteration

	class genasync:

		def __init__(self):
			self.i = 3

		def __aiter__(self):
			return self

		@asyncio.coroutine
		def __anext__(self):
			if self.i > 0:
				self.i -= 1
				return "asd"
			else:
				raise StopAsyncIteration

	for i in progress_content(gensync()):
		pass

	print()

	@asyncio.coroutine
	async def run():
		async for i in progress_content(genasync()):
			pass

	loop = asyncio.get_event_loop()
	loop.run_until_complete(run())
	loop.close()



from __future__ import absolute_import, division, print_function, unicode_literals

from past.builtins import cmp

from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import MutableSequence, Callable

def _insertion(seq, cmp_, left, right, gap):
	# type: (MutableSequence, Callable, int, int, int) -> None

	loc = left + gap
	while loc <= right:
		i = loc - gap
		value = seq[loc]
		while i >= left and cmp_(seq[i], value) > 0:
			seq[i + gap] = seq[i]
			i -= gap
		seq[i+gap] = value
		loc += gap

GROUP_SIZE = 5

def median_of_medians(seq, cmp_=None, left=0, right=None, depth=0):
	# type: (MutableSequence, Callable, int, int, int)

	""" Approximate median selection algorithm.
		This is not the full median of medians algorithm.
		Currently only works for len(seq) == GROUP_SIZE ** x
	"""

	offset = (GROUP_SIZE ** (depth + 1) - GROUP_SIZE ** depth) // 2

	gap = GROUP_SIZE ** depth

	cmp_ = cmp_ or cmp

	if not right:
		right = len(seq) - 1

	span = GROUP_SIZE*gap
	num = (right - left + 1) // span

	if num == 0:
		_insertion(seq, cmp_, left, right, gap)
		num = (right - left + 1) // gap
		return left + gap * (num - 1) // 2  # select median

	s = left
	while s < right:
		_insertion(seq, cmp_, s, s + span - 1, gap)
		s += span

	if num < GROUP_SIZE:
		_insertion(seq, cmp_, left + offset, right, span)
		return left + offset + num * span // 2  # select median
	else:
		return median_of_medians(seq, cmp_, left + offset, right, depth + 1)



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewitems
import logging
from functools import wraps
from collections import defaultdict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Callable, Collection, Optional, Any

def _to_union(types):
	# type: (Collection[str], ) -> str

	types = list(set(types))

	if not types:
		return ""
	elif len(types) == 1:
		return types[0]
	elif len(types) > 1:
		return "Union[" + ", ".join(types) + "]"

def _type_str(obj):
	# type: (Any, ) -> str

	return type(obj).__name__

def rec_repr(obj):
	# type: (Any, ) -> str

	if isinstance(obj, defaultdict):
		return _type_str(obj) + "[" + rec_repr(obj.default_factory) + "]"
	elif isinstance(obj, list):
		return "List[" + _to_union(map(rec_repr, obj)) + "]"
	elif isinstance(obj, set):
		return "Set[" + _to_union(map(rec_repr, obj)) + "]"
	else:
		return _type_str(obj)

def _arg_str(arg, maxlen=None, app="...", repr_args=True):
	# type: (Any, Optional[int], bool) -> str

	if repr_args:
		arg = repr(arg)

	assert isinstance(arg, str)

	if maxlen:
		if len(arg) <= maxlen + len(app):
			return arg
		else:
			return arg[:maxlen] + app
	else:
		return arg

def _kwarg_str(key, value, maxlen=None, app="...", repr_args=True):
	# type: (str, Any, Optional[int], str, bool) -> str

	return key + "=" + _arg_str(value, maxlen, app, repr_args)

def args_str(args, kwargs, maxlen=20, app="...", repr_args=True):
	# type: (tuple, dict, Optional[int], str, bool) -> str

	""" Creates printable string from function arguments. """

	args = ", ".join(_arg_str(arg, maxlen, app, repr_args) for arg in args)
	kwargs = ", ".join(_kwarg_str(k, v, maxlen, app, repr_args) for k, v in viewitems(kwargs))

	if args:
		if kwargs:
			return args + ", " + kwargs
		else:
			return args
	else:
		if kwargs:
			return kwargs
		else:
			return ""

def log_call(s):
	# type: (str, ) -> Callable

	""" Decorator to log function calls using template string `s`.
		Available format fields are: 'name', 'args' and 'kwargs'.
	"""

	def dec(func):
		# type: (Callable, ) -> Callable

		def inner(*args, **kwargs):
			logging.debug(s.format(
				name=func.__name__,
				args=args,
				kwargs=kwargs
			))
			return func(*args, **kwargs)
		return inner
	return dec

def log_wrap_call(func):
	# type: (Callable, ) -> Callable

	""" Decorator which logs all calls to `func` with all arguments. """

	@wraps(func)
	def inner(*args, **kwargs):
		logging.debug("START %s(%s)", func.__name__, args_str(args, kwargs))

		try:
			ret = func(*args, **kwargs)
		except BaseException as e:
			logging.exception("RAISED %s(%s)", func.__name__, args_str(args, kwargs))
			raise

		logging.debug("END %s(%s)", func.__name__, args_str(args, kwargs))
		return ret
	return inner

def log_methodcall(func):
	# type: (Callable, ) -> Callable

	""" Decorator to log method calls with arguments. """

	@wraps(func)
	def inner(self, *args, **kwargs):
		classname = self.__class__.__name__
		# classname = type(self).__name__ ?
		logging.debug("%s.%s(%s)", classname, func.__name__, args_str(args, kwargs))
		return func(self, *args, **kwargs)
	return inner

def log_methodcall_result(func):
	# type: (Callable, ) -> Callable

	""" Decorator to log method calls with arguments and results. """

	@wraps(func)
	def inner(self, *args, **kwargs):
		classname = self.__class__.__name__
		# classname = type(self).__name__ ?
		logging.debug("%s.%s(%s)", classname, func.__name__, args_str(args, kwargs))
		res = func(self, *args, **kwargs)
		logging.debug("%s.%s => %s", classname, func.__name__, res)
		return res
	return inner



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import input, range

import os, sys, os.path, shlex
from functools import wraps
from argparse import ArgumentTypeError
from typing import TYPE_CHECKING

from .iter import is_empty
from .stdio import confirm
from .compat.pathlib import Path
from .compat.os import makedirs

if TYPE_CHECKING:
	from typing import Callable

def get_args(argparser):
	""" get commandline arguments from std input instead """

	from pprint import pprint

	if len(sys.argv) > 1:
		return argparser.parse_args()

	print("stdin")
	args = []

	for action in argparser._actions:
		pprint(action)

	for k, v in argparser._option_string_actions.items():

		nargs = v.nargs if v.nargs is not None else 1

		if nargs == 0:
			if confirm(str(k)):
				args.append(k)
		else:
			instr = input("{} ({}): ".format(k, nargs))  # separate multiple value by whitespace, quoting supported.
			args.append(k)
			args.append(instr)

	args = shlex.split(" ".join(args))
	pprint(args)
	return argparser.parse_args(args)

def arg_to_path(func):
	# type: (Callable[[Path], Path], ) -> Callable

	@wraps(func)
	def inner(path):
		return func(Path(path))
	return inner

def multiple_of(divisor):
	# type: (int, ) -> Callable[[str], int]

	from builtins import int as builtin_int

	""" This function is called 'int' so that argparse can show a nicer error message
		in case input cannot be cast to int:
		error: argument --multiple: invalid int value: 'a'
	"""
	def int(s):
		# type: (str, ) -> int

		number = builtin_int(s)

		if number % divisor != 0:
			msg = "{0} is not clearly divisible by {1}".format(s, divisor)
			raise ArgumentTypeError(msg)

		return number

	return int

def in_range(start, stop, step=1):
	# type: (int, int, int) -> Callable[[str], int]

	from builtins import int as builtin_int

	def int(s):  # see: multiple_of()
		# type: (str, ) -> int

		number = builtin_int(s)

		r = range(start, stop, step)
		if number not in r:
			msg = "{0} is not in {1}".format(s, r)
			raise ArgumentTypeError(msg)

		return number

	return int

def between(start, stop):
	# type: (float, float) -> Callable[[str], float]

	from builtins import float as builtin_float

	def float(s):
		# type: (str, ) -> float

		number = builtin_float(s)

		if not (start <= number < stop):
			msg = "{0} is not in between {1} and {2}".format(s, start, stop)
			raise ArgumentTypeError(msg)

		return number

	return float

def suffix(s):
	# type: (str, ) -> str

	""" Checks if `s` is a valid suffix. """

	if not s.startswith("."):
		msg = "{0} is not a valid suffix. It must start with a dot.".format(s)
		raise ArgumentTypeError(msg)

	return s

def lowercase(s):
	# type: (str, ) -> str

	""" Converts argument to lowercase. """

	return s.lower()

@arg_to_path
def existing_path(path):
	# type: (Path, ) -> Path

	""" Checks if a path exists. """

	if not path.exists():
		msg = "{0} does not exist".format(path)
		raise ArgumentTypeError(msg)

	return path

@arg_to_path
def new_path(path):
	# type: (Path, ) -> Path

	""" Checks if a path exists. """

	if path.exists():
		msg = "{0} already exists".format(path)
		raise ArgumentTypeError(msg)

	return path

@arg_to_path
def is_dir(path):
	# type: (Path, ) -> Path

	"""Checks if a path is an actual directory"""

	if not path.is_dir():
		msg = "{0} is not a directory".format(path)
		raise ArgumentTypeError(msg)

	return path

@arg_to_path
def abs_path(path):
	# type: (Path, ) -> Path

	"""Checks if a path is an actual directory"""

	return path.resolve()

@arg_to_path
def is_file(path):
	# type: (Path, ) -> Path

	"""Checks if a path is an actual file"""

	if not path.is_file():
		msg = "{0} is not a file".format(path)
		raise ArgumentTypeError(msg)

	return path

@arg_to_path
def future_file(path):
	# type: (Path, ) -> Path

	""" Tests if file can be created to catch errors early.
		Checks if directory is writeable and file does not exist yet.
	"""

	if path.parent and not os.access(str(path.parent), os.W_OK):
		msg = "cannot access directory {0}".format(path.parent)
		raise ArgumentTypeError(msg)
	if path.is_file():
		msg = "file {0} already exists".format(path)
		raise ArgumentTypeError(msg)
	return path

@arg_to_path
def out_dir(path):
	# type: (Path, ) -> Path

	""" Tests if `path` is a directory. If not it tries to create one.
	"""

	if not path.is_dir():
		try:
			makedirs(path)
		except OSError:
			msg = "Error: '{}' is not a valid directory.".format(path)
			raise ArgumentTypeError(msg)

	return path

@arg_to_path
def empty_dir(dirname):
	# type: (Path, ) -> Path

	""" tests if directory is empty """

	if not is_empty(os.scandir(dirname)):
		msg = "directory {0} is not empty".format(dirname)
		raise ArgumentTypeError(msg)

	return dirname

if __name__ == "__main__":

	from argparse import ArgumentParser

	parser = ArgumentParser()
	parser.add_argument('--str')
	parser.add_argument('--required', action='store_true')
	parser.add_argument('--add', nargs="+")
	print(get_args(parser))

	"""
	parser = ArgumentParser()
	parser.add_argument('--indir', type=is_dir)
	parser.add_argument('--infile', type=is_file)
	parser.add_argument('--outfile', type=future_file)
	parser.add_argument('--outdir', type=empty_dir)
	args = parser.parse_args()
	print(args)
	"""



from __future__ import absolute_import, division, print_function, unicode_literals

import logging, time
from datetime import tzinfo, timedelta

from .compat.datetime import datetime, timezone

utcmin = datetime.min.replace(tzinfo=timezone.utc)
utcmax = datetime.max.replace(tzinfo=timezone.utc)

def now():
	# type: () -> datetime

	""" Returns the current datetime as timezone aware object in UTC timezone. """

	return datetime.now(timezone.utc)

# was: datetime_from_utc_timestamp
def datetime_from_utc_timestamp(epoch):
	# type: (int, ) -> datetime

	""" Converts a UNIX epoch time to a timezone aware datetime. """

	return datetime.fromtimestamp(epoch, timezone.utc)

class LocalTimezone(tzinfo):

	def __init__(self):
		if time.daylight:
			self.daylight = timedelta(seconds=-time.altzone) - timedelta(seconds=-time.timezone)
			self.offset = timedelta(seconds=-time.altzone)
			self.name = time.tzname[1]
		else:
			self.daylight = timedelta(0)
			self.offset = timedelta(seconds=-time.timezone)
			self.name = time.tzname[0]

	def dst(self, dt):
		return self.daylight

	def utcoffset(self, dt):
		return self.offset

	def tzname(self):
		return self.name

	def __repr__(self):
		return "genutility.datetime.localtimezone"

localtimezone = LocalTimezone()

def localnow():
	# type: () -> datetime

	""" Returns the current datetime as timezone aware object in local timezone. """

	return datetime.now(localtimezone)



from __future__ import absolute_import, division, print_function, unicode_literals

import pickle  # nosec
from pickle import HIGHEST_PROTOCOL  # nosec
from functools import wraps
from datetime import timedelta
from typing import TYPE_CHECKING

from .atomic import TransactionalCreateFile
from .compat import FileNotFoundError
from .compat.contextlib import nullcontext
from .compat.os import fspath
from .datetime import now
from .file import copen
from .filesystem import mdatetime
from .object import args_to_key
from .time import PrintStatementTime

if TYPE_CHECKING:
	from typing import Any, Callable, Optional, Iterator, Iterable
	from .compat.pathlib import Path

def read_pickle(path):
	# type: (str, ) -> Any

	""" Read pickle file from `path`.
		Warning: All usual security consideration regarding the pickle module still apply.
	"""

	with copen(path, "rb") as fr:
		return pickle.load(fr)  # nosec

def write_pickle(result, path, protocol=None, safe=False):
	# type: (Any, str, Optional[int], bool) -> None

	""" Write `result` to `path` using pickle serialization.

		`protocol': pickle protocol version
		`safe`: if True, don't overwrite original file in case any error occurs
	"""

	if safe:
		context = TransactionalCreateFile
	else:
		context = copen

	with context(path, "wb") as fw:
		pickle.dump(result, fw, protocol=protocol)

def read_iter(path):
	# type: (str, ) -> Iterator[Any]

	""" Read pickled iterable from `path`. 
		Warning: All usual security consideration regarding the pickle module still apply.
	"""

	with copen(path, "rb") as fr:
		unpickler = pickle.Unpickler(fr)  # nosec
		while fr.peek(1):
			yield unpickler.load()

def write_iter(it, path, protocol=None, safe=False):
	# type: (Iterable[Any], str, Optional[int], bool) -> Iterator[Any]

	""" Write iterable `it` to `path` using pickle serialization. This uses much less memory than
			writing a full list at once.
		Read back using `read_iter()`. If `safe` is True, the original file is not overwritten
			if any error occurs.
		This is a generator which yields the values read from `it`. So it must be consumed
			to actually write anything to disk.
	"""

	if safe:
		context = TransactionalCreateFile
	else:
		context = copen

	with context(path, "wb") as fw:
		pickler = pickle.Pickler(fw, protocol=protocol)
		for result in it:
			pickler.dump(result)
			yield result

def key_to_hash(key, protocol=None):
	from hashlib import md5
	binary = pickle.dumps(key, protocol=protocol)
	return md5(binary).hexdigest()  # nosec

def cache(path, duration=None, generator=False, protocol=HIGHEST_PROTOCOL, ignoreargs=False, verbose=False):
	# type: (Path, Optional[timedelta], bool, int, bool, bool) -> Callable[[Callable], Callable]

	""" Decorator to cache function calls. Doesn't take function arguments into regard.
		It's using `pickle` to deserialize the data. So don't use it with untrusted inputs.

		`path`: path to cache file
		`duration`: maximum age of cache
		`generator`: set to True to store the results of generator objects
		`protocol`: pickle protocol version
		`verbose`: print loading times if `generator` is False

		If `ignoreargs` is True, the cache won't take function arguments into regard.
			The path will be interpreted as template string to a file instead of a directory
			where `ppv` is assigned the  pickle protocol version.
	"""

	duration = duration or timedelta.max

	if not verbose:
		context = nullcontext
	else:
		context = PrintStatementTime

	def decorator(func):
		# type: (Callable, ) -> Callable

		@wraps(func)
		def inner(*args, **kwargs):
			# type: (*Any, **Any) -> Any

			if not ignoreargs:
				hash = key_to_hash(args_to_key(args, kwargs), protocol=protocol)
				fullpath = path / hash
			else:
				if args or kwargs:
					logger.warning("cache file decorator for %s called with arguments", func.__name__)
				fullpath = fspath(path).format(ppv=protocol)

			try:
				invalid = now() - mdatetime(fullpath) > duration
			except FileNotFoundError:
				invalid = True

			if invalid:
				if not ignoreargs:
					path.mkdir(parents=True, exist_ok=True)

				if generator:
					it = func(*args, **kwargs)
					return write_iter(it, fullpath, protocol=protocol, safe=True)
				else:
					with context("Result calculated in {delta} seconds"):
						result = func(*args, **kwargs)
					write_pickle(result, fullpath, protocol=protocol, safe=True)
					return result
			else:
				if generator:
					return read_iter(fullpath)
				else:
					with context("Result loaded from cache in {delta} seconds"):
						return read_pickle(fullpath)
		return inner

	return decorator



from __future__ import absolute_import, division, print_function, unicode_literals

from io import open
from typing import TYPE_CHECKING

from PyPDF2 import PdfFileMerger, PdfFileReader

from .compat import FileNotFoundError
from .compat.contextlib import ExitStack
from .compat.pathlib import Path

if TYPE_CHECKING:
	from typing import Union, Iterator
	from .compat.os import PathLike

def join_pdfs_in_folder(path_in, file_out, overwrite=False):
	# type: (Path, Union[PathLike, str]) -> None

	""" Join all PDFs in `path_in` and write to `file_out`. """

	merger = PdfFileMerger(strict=True)

	if not path_in.exists():
		raise FileNotFoundError(path_in)

	with ExitStack() as stack:
		for path in path_in.glob("*.pdf"):
			if path.is_file():
				fr = stack.enter_context(path.open("rb")) # open does not support Path in <3.6
				merger.append(fr)

		if overwrite:
			mode = "w"
		else:
			mode = "x"

		with open(file_out, mode+"b") as fw:
			merger.write(fw)

def iter_pdf_text(path):
	# type: (str, ) -> Iterator[str]

	with open(path, "rb") as fr:
		pdf = PdfFileReader(fr)
		for i in range(pdf.getNumPages()):
			page = pdf.getPage(i)
			yield page.extractText()

def _read_pdf_pdfminer(path):
	from pdfminer.high_level import extract_text

	return extract_text(path)

def _read_pdf_tika(path):
	from tika import parser

	raw = parser.from_file(path)

	#return raw["metadata"]
	return raw["content"]

def read_pdf(path, engine="pdfminer"):
	# type: (str, str) -> str

	try:
		func = {
			"pdfminer": _read_pdf_pdfminer,
			"tika": _read_pdf_tika,
		}[engine]
	except KeyError:
		raise ValueError("Engine {} doesn't exist".format(engine))

	return func(path)

if __name__ == "__main__":
	from argparse import ArgumentParser

	parser = ArgumentParser(description="Merge pdf files in directory into one file.")
	parser.add_argument("dir", help="input directory")
	parser.add_argument("out", help="output file path")
	args = parser.parse_args()

	join_pdfs_in_folder(args.dir, args.out)



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from typing import Set, TypeVar
	T = TypeVar("T")

def get(s):
	# type: (Set[T], ) -> T

	for i in s:
		return i



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import TYPE_CHECKING

import cv2
import numpy as np

if TYPE_CHECKING:
	from typing import Union, Iterator
	import wx

logger = logging.getLogger(__name__)

def wx_to_cv_image(wximage):
	# type (wx.Image, ) -> np.ndarray

	buf = wximage.GetDataBuffer()
	arr = np.asarray(buf, dtype=np.uint8)

	return np.reshape(arr, (wximage.Height, wximage.Width, 3))[:,:,::-1]  # RGB => BGR

def grayscale(cvimg):
	# type: (np.ndarray, ) -> np.ndarray

	return cv2.cvtColor(cvimg, cv2.COLOR_RGB2GRAY)

def iter_video(input=0, show=False):
	# type: (Union[str, int], bool) -> Iterator[np.ndarray]

	cap = cv2.VideoCapture(input)
	logger.debug("Reading video using %s backend", cap.getBackendName())

	try:
		while True:
			retval, image = cap.read()
			if retval:
				yield image
			else:
				break

			if show:
				cv2.imshow("iter_video", image)
				if cv2.waitKey(1) & 0xFF == ord('q'):
					return
	finally:
		cap.release()
		if show:
			cv2.destroyAllWindows()



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewitems, viewvalues

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
	from typing import Any, Callable, Dict, List, Iterator, Iterable, Optional, Set, Tuple
	from graphviz import Digraph

if __debug__:
	import graphviz

class BKNode(object):

	__slots__ = ["value", "leaves"]

	def __init__(self, value, leaves):
		# type: (Any, Dict[Any, BKNode]) -> None

		self.value = value
		self.leaves = leaves

	def __str__(self):
		return str(self.value) + ": " + str(self.leaves)

	def __repr__(self):
		return repr(self.value) + ": " + repr(self.leaves)

class BKTree(object):

	""" BK-tree. Used for nearest neighbor queries according to a discrete metric.
		Examples for the metric are the Manhattan distance or the Levenshtein distance.
	"""

	def __init__(self, distance_func):
		# type: (Callable[[Any, Any], int], ) -> None

		self.distance_func = distance_func
		self.root = None # type: Optional[BKNode]

	def add(self, value):
		# type: (Any, ) -> None

		if self.root is None:
			self.root = BKNode(value, {})
			return

		node = self.root
		while True:
			distance = self.distance_func(value, node.value)
			try:
				node = node.leaves[distance]
			except KeyError:
				node.leaves[distance] = BKNode(value, {})
				break

	def update(self, values):
		# type: (Iterable[Any], ) -> None

		for value in values:
			self.add(value)

	def find(self, value, max_distance):
		# type: (Any, int) -> List[Tuple[int, Any]]

		""" Returns all values from tree where the metric distance
			is less or equal to `max_distance`.
		"""

		node = self.root
		ret = [] # type: List[Tuple[int, Any]]

		if node is None:
			return ret

		candidates = [node] # is a deque better here?

		while candidates:
			candidate = candidates.pop()
			distance = self.distance_func(value, candidate.value)

			if distance <= max_distance:
				ret.append((distance, candidate.value))

			# instead of looking for candidates by searching,
			# one could also directly access the necessary keys in the dict
			for d, bknode in viewitems(candidate.leaves):
				lower = distance - max_distance
				upper = distance + max_distance
				if lower <= d <= upper:
					candidates.append(bknode)

		return ret

	@staticmethod
	def _find_by_distance(node, distance):
		# type: (BKNode, int) -> Iterator[Set[Any]]

		for d, bknode in viewitems(node.leaves):
			if d == distance:
				nodeset = set(BKTree._values(bknode))
				nodeset.add(node.value)
				yield nodeset
			for ret in BKTree._find_by_distance(bknode, distance):
				yield ret

	def find_by_distance(self, distance):
		# type: (int, ) -> Iterator[Set[Any]]

		""" Find all sets of values base on `distance` between each other.
		"""

		if self.root is None:
			return iter([])
		else:
			return BKTree._find_by_distance(self.root, distance)

	@staticmethod
	def _dot(dot, node):
		# type: (Digraph, BKNode) -> None

		for distance, childnode in viewitems(node.leaves):
			dot.node(str(childnode.value))
			dot.edge(str(node.value), str(childnode.value), label=str(distance))
			BKTree._dot(dot, childnode)

	def saveimage(self, filename, format="png"):
		from graphviz import Digraph

		if self.root is None:
			raise ValueError("Tree is empty")

		dot = Digraph(format=format)
		dot.node(str(self.root.value))
		self._dot(dot, self.root)
		dot.render(filename)

	@staticmethod
	def _values(node):
		# type: (BKNode, ) -> Iterator[Any]

		yield node.value
		for leaf in viewvalues(node.leaves):
			for value in BKTree._values(leaf):
				yield value

	def values(self):
		# type: () -> Iterator[Any]

		if self.root is None:
			return iter([])
		else:
			return self._values(self.root)

	def __iter__(self):
		# type: () -> Iterator[Any]

		return self.values()

if __name__ == "__main__":
	from Levenshtein import distance as levenshtein_distance # pip install python-Levenshtein
	#from leven import levenshtein as levenshtein_distance # pip install leven

	from argparse import ArgumentParser
	parser = ArgumentParser()
	parser.add_argument("outpath", default="bk-tree.gv", nargs="?")
	args = parser.parse_args()

	words = ["laptop", "security", "microsoft", "computer", "software", "tree", "algorithm", "desktop"]

	tree = BKTree(levenshtein_distance)
	tree.update(words)
	tree.saveimage(args.outpath)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import TYPE_CHECKING

import numpy as np
from PIL import ImageFilter

#from .numba import opjit
from .numpy import unblock, rgb_to_hsi, rgb_to_ycbcr

# fingerprinting aka perceptual hashing

if TYPE_CHECKING:
	from PIL import Image

def phash_antijpeg(image):
	# type: (Image, ) -> np.ndarray

	""" Source: An Anti-JPEG Compression Image Perceptual Hashing Algorithm
		`image` is a RGB pillow image.
	"""

	raise NotImplementedError

def hu_moments(channels):

	""" Calculates all Hu invariant image moments for all channels separately.
		Input array must be of shape [width, height, channels]
		Returns shape [moments, channels]
	"""

	# pre-calculate matrices
	n, m, _ = channels.shape
	coords_x, coords_y = np.meshgrid(np.arange(m), np.arange(n))
	coords_x = np.expand_dims(coords_x, axis=-1) # for batch input, some change is needed here
	coords_y = np.expand_dims(coords_y, axis=-1) # for batch input, some change is needed here

	def M(p, q):
		return np.sum(coords_x**p * coords_y**q * channels, axis=(-2, -3))

	def mu(p, q, xb, yb):
		return np.sum((coords_x - xb)**p * (coords_y - yb)**q * channels, axis=(-2, -3))

	def eta(p, q, xb, yb, mu00):
		gamma = (p+q)/2 + 1
		return mu(p, q, xb, yb) / mu00**gamma

	def loop():

		M00 = M(0, 0)
		if not np.all(M00 > 0.):
			logging.error("M00: %s", M00)
			raise ValueError("Failed to calculate moments. Single color pictures are not supported yet.")

		M10 = M(1, 0)
		M01 = M(0, 1)
		xb = M10 / M00
		yb = M01 / M00

		mu00 = mu(0, 0, xb, yb)

		eta20 = eta(2, 0, xb, yb, mu00)
		eta02 = eta(0, 2, xb, yb, mu00)
		eta11 = eta(1, 1, xb, yb, mu00)
		eta30 = eta(3, 0, xb, yb, mu00)
		eta12 = eta(1, 2, xb, yb, mu00)
		eta21 = eta(2, 1, xb, yb, mu00)
		eta03 = eta(0, 3, xb, yb, mu00)

		phi1 = eta20 + eta02
		phi2 = (eta20 - eta02)**2 + 4*eta11**2
		phi3 = (eta30-3*eta12)**2 + (3*eta21 - eta03)**2
		phi4 = (eta30 + eta12)**2 + (eta21 + eta03)**2
		phi5 = (eta30 - 3*eta12)*(eta30+eta12)*((eta30+eta12)**2 - 3*(eta21 + eta03)**2) + (3*eta21 - eta03)*(eta21 + eta03)*(3*(eta30 + eta12)**2 - (eta21 + eta03)**2)
		phi6 = (eta20 - eta02)*((eta30 + eta12)**2 - (eta21 + eta03)**2) + 4*eta11*(eta30+eta12)*(eta21 + eta03)
		phi7 = (3*eta21 - eta03)*(eta30 + eta12)*((eta30 + eta12)**2 - 3*(eta21 + eta03)**2) - (eta30 - 3*eta12)*(eta21 + eta03)*(3*(eta30 + eta12)**2 - (eta21 + eta03)**2)

		return np.array([phi1, phi2, phi3, phi4, phi5, phi6, phi7])

	return loop()

#@opjit() rgb_to_hsi and rgb_to_ycbcr not supported by numba
def phash_moments_array(arr):
	arr = arr / 255.

	# convert colorspaces
	hsi = rgb_to_hsi(arr)
	ycbcr = rgb_to_ycbcr(arr)#.astype(np.uint8)
	channels = np.concatenate([hsi, ycbcr], axis=-1)
	return np.concatenate(hu_moments(channels).T)

def phash_moments(image):
	# type: (Image, ) -> np.ndarray

	""" Source: Perceptual Hashing for Color Images Using Invariant Moments
		`image` is a RGB pillow image. Results should be compared with L^2-Norm of difference vector.
	"""

	if image.mode != "RGB":
		raise ValueError("Only RGB images are supported")

	# preprocessing
	image = image.resize((512, 512), Image.BICUBIC)
	image = image.filter(ImageFilter.GaussianBlur(3))
	image = np.array(image)
	return phash_moments_array(image)

def phash_blockmean_array(arr, bits=256):
	# type: (np.ndarray, int) -> np.ndarray

	""" If bits is not a multiple of 8,
		the result will be zero padded from the right.
	"""

	if len(arr.shape) != 2:
		raise ValueError("arr must be 2-dimensional")

	n = int(np.sqrt(bits))
	if n**2 != bits:
		raise ValueError("bits must be a square number")

	blocks = unblock(arr, n, n)
	means = np.mean(blocks, axis=-1)
	median = np.median(means)
	bools = means >= median
	return np.packbits(bools)

def phash_blockmean(image, bits=256, x=256):
	# type: (Image, int, int) -> bytes

	""" Source: Block Mean Value Based Image Perceptual Hashing
		Method: 1
		Metric: 'Bit error rate' (normalized hamming distance)
	"""

	image = image.convert('L').resize((x, x))
	image = np.array(image)
	return phash_blockmean_array(image, bits).tobytes()



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import chr, zip
from future.moves.itertools import zip_longest

import os, platform, stat, os.path, re, logging, shutil, errno
from operator import attrgetter
from fnmatch import fnmatch
from functools import partial
from typing import TYPE_CHECKING

from .file import equal_files, iterfilelike, FILE_IO_BUFFER_SIZE
from .iter import is_empty
from .os import islink, uncabspath, _not_available
from .ops import logical_implication
from .datetime import datetime_from_utc_timestamp
from .compat import FileExistsError
from .compat.os import PathLike, fspath, scandir
from .compat.pathlib import Path

if TYPE_CHECKING:
	from typing import Any, Callable, Optional, Iterator, Iterable, List, Set, Tuple, Union
	from datetime import datetime
	PathType = Union[str, PathLike]
	from .compat.os import DirEntry
	EntryType = Union[Path, DirEntry]

logger = logging.getLogger(__name__)

ascii_control = set(chr(i) for i in range(1, 32))
ntfs_illegal_chars = {"\0", "/"}
fat_illegal_chars = {"\0", "/", "?", "<", ">", "\\", ":", "*", "|", '"', "^"} # ^ really?

windows_ntfs_illegal_chars = ntfs_illegal_chars | {"\\", ":", "*", "?", "\"", "<", ">", "|"} | ascii_control
windows_fat_illegal_chars = fat_illegal_chars

windows_illegal_chars = windows_ntfs_illegal_chars | windows_fat_illegal_chars
unix_illegal_chars = {"\0", "/"}
mac_illegal_chars = {"\0", "/", ":"}

windows_sep = r"\/"
linux_sep = r"/"
mac_sep = r"/:"

windows_reserved_names = {"CON", "PRN", "AUX", "NUL", "COM1", "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9", "LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", "LPT9"}

class fileextensions(object):
	audio = ("wav", "mp3", "aac", "flac", "m4a", "m4b", "aiff", "ogg", "wma", "mka", "ac3", "dts")
	video = ("avi", "mp4", "m4v", "mkv", "mpg", "mpeg", "wmv", "mov", "flv", "f4v", "webm", "vob", "ogv", "m2ts", "rm", "rmvb", "asf", "3gp", "nsv", "divx", "ogm")
	images = ("bmp", "jpg", "jpeg", "png", "gif", "tif", "tiff", "tga")

	archives = ("zip", "rar", "ace", "tar", "tgz", "tbz", "7z", "cab", "dmg", "wim")
	executables = ("exe", "dll", "so", "scr", "msi", "pyd", "sys")
	scripts = ("bat", "ps1", "sh", "py", "php")
	documents = ("txt", "pdf", "doc", "docx", "ps")
	image_archives = ("cbz", "cbr", "cb7", "cbt", "cba")

	disc_images_binary = ("iso", "bin", "mdf", "img")
	disc_images_sidecar = ("cue", "mds", "ccd")
	disc_images = disc_images_binary + disc_images_sidecar

	game_images = ("gcn", "nds", "3ds", "wad", "xbx", "gba", "gb", "nes", "sfc", "n64", "z64")
	compressed = ("gz", "bz2", "lzma", "xz", "z")
	subtitles = ("srt", "sub", "idx", "sup", "ssa", "ass")

	scene = ("nfo", "sfv")

class IllegalFilename(ValueError):
	pass

class WindowsIllegalFilename(IllegalFilename):
	pass

class FileProperties(object):

	__slots__ = ("relpath", "size", "isdir", "abspath", "id", "modtime", "hash")

	def __init__(self, relpath, size, isdir, abspath=None, id=None, modtime=None, hash=None):
		# type: (str, int, bool, Optional[str], Optional[Any], Optional[datetime], Optional[str]) -> None

		self.relpath = relpath
		self.size = size
		self.isdir = isdir
		self.abspath = abspath
		self.id = id
		self.modtime = modtime
		self.hash = hash

	@classmethod
	def keys(cls):
		return cls.__slots__

	def values(self):
		return attrgetter(*self.__slots__)(self)

	def __iter__(self):
		# type: () -> tuple

		return self.values()

	def __repr__(self):
		# type: () -> str

		args = ("{}={!r}".format(k, v) for k, v in zip(self.keys(), self.values.values()) if v is not None)
		return "FileProperties({})".format(", ".join(args))

class DirEntryStub(object):
	__slots__ = ("name", "path")

	def __init__(self, name, path):
		self.name = name
		self.path = path

class BaseDirEntry(object):
	__slots__ = ("entry", )

	def __init__(self, entry):
		self.entry = entry

	@property
	def name(self):
		return self.entry.name

	@property
	def path(self):
		return self.entry.path

	def inode(self):
		return self.entry.inode()

	def is_dir(self):
		return self.entry.is_dir()

	def is_file(self):
		return self.entry.is_file()

	def is_symlink(self):
		return self.entry.is_symlink()

	def stat(self):
		return self.entry.stat()

	def __str__(self):
		return str(self.entry)

	def __repr__(self):
		return repr(self.entry)

	def __fspath__(self):
		return self.entry.path

class MyDirEntry(BaseDirEntry):
	__slots__ = ("basepath", "_relpath", "follow")

	def __init__(self, entry):
		BaseDirEntry.__init__(self, entry)

		self.basepath = None # type: Optional[str]
		self._relpath = None # type: Optional[str]
		self.follow = True

	@property
	def relpath(self):
		# type: () -> str

		if self._relpath is not None:
			return self._relpath

		if self.basepath is None:
			raise RuntimeError("relpath cannot be returned, because basepath is not set")

		self._relpath = os.path.relpath(self.entry.path, self.basepath)
		return self._relpath

	@relpath.setter
	def relpath(self, path):
		# type: (str, ) -> None

		self._relpath = path

if TYPE_CHECKING:
	MyDirEntryT = Union[DirEntry, MyDirEntry]

def mdatetime(path):
	# type: (PathType, ) -> datetime

	""" Returns the last modified date of `path`. """

	if isinstance(path, Path):
		mtime = path.stat().st_mtime
	else:
		mtime = os.stat(path).st_mtime

	return datetime_from_utc_timestamp(mtime)

def rename(old, new):
	# type: (PathType, PathType) -> None

	""" Renames `old` to `new`. Fails if `new` already exists.
		This is the default behaviour of `os.rename` on Windows.
		This function should do the same cross-platform.
		It is however not race free.
	"""

	# fixme: renaming a non-existing file in a non-existing folder yields a PermissionError not FileNotFoundError

	if os.path.exists(new):
		raise FileExistsError(new)

	os.renames(old, new) # fixme: don't use rename*s*

def copy_file_generator(source, dest, buffer_size=FILE_IO_BUFFER_SIZE, overwrite_readonly=False):
	# type: (PathType, PathType, int, bool) -> Iterator[None]

	""" Partial file is not deleted if exception gets raised anywhere. """

	try:
		os.makedirs(os.path.dirname(dest))
	except OSError as e:
		if e.errno != errno.EEXIST:
			raise

	if overwrite_readonly:
		try:
			make_writeable(dest)
		except OSError as e:
			if e.errno != errno.ENOENT:
				raise

	with open(source, "rb") as src, open(dest, "wb") as dst:
		for data in iterfilelike(src, chunk_size=buffer_size):
			dst.write(data)
			yield

	shutil.copystat(source, dest)

def st_mode_to_str(st_mode):
	# type: (int, ) -> str

	if stat.S_ISDIR(st_mode): return "directory"
	elif stat.S_ISREG(st_mode): return "regular file"
	elif stat.S_ISCHR(st_mode): return "character special device file"
	elif stat.S_ISBLK(st_mode): return "block special device file"
	elif stat.S_ISFIFO(st_mode): return "named pipe"
	elif stat.S_ISLNK(st_mode): return "symbolic link"
	elif stat.S_ISSOCK(st_mode): return "socket"
	else: return "unknown"

def append_to_filename(path, s):
	# type: (PathType, str) -> str

	""" Add `s` to the filename given in `path`. It's added before the extension, not after. """

	root, ext = os.path.splitext(path)
	return root + s + ext

def scandir_error_log(entry, exception):
	logger.exception("Error in %s", entry.path, exc_info=exception)

def scandir_error_raise(entry, exception):
	raise exception

def scandir_error_ignore(entry, exception):
	pass

def _scandir_rec_skippable(rootentry, files=True, others=False, follow_symlinks=True, errorfunc=scandir_error_raise):
	# type: (DirEntry, bool, bool, bool, Callable[[DirEntry, Exception], None]) -> Iterator[MyDirEntry]

	try:
		with scandir(rootentry.path) as it: # python 3.6
			for entry in it:
				if files and entry.is_file(follow_symlinks=follow_symlinks):
					yield MyDirEntry(entry)
				elif entry.is_dir(follow_symlinks=follow_symlinks):
					if not follow_symlinks and islink(entry): # must be a windows junction
						continue
					se = MyDirEntry(entry)
					yield se
					if se.follow:
						for e in _scandir_rec_skippable(entry, files, others, follow_symlinks, errorfunc):
							yield e
				else:
					if others:
						yield MyDirEntry(entry)
	except OSError as e:
		errorfunc(rootentry, e)

def _scandir_rec(rootentry, files=True, dirs=False, others=False, rec=True, follow_symlinks=True, errorfunc=scandir_error_raise):
	# type: (DirEntry, bool, bool, bool, bool, bool, Callable[[DirEntry, Exception], None]) -> Iterator[DirEntry]

	try:
		with scandir(rootentry.path) as it: # python 3.6
			for entry in it:
				if files and entry.is_file(follow_symlinks=follow_symlinks):
					yield entry
				elif entry.is_dir(follow_symlinks=follow_symlinks):
					if not follow_symlinks and islink(entry): # must be a windows junction
						continue
					if dirs:
						yield entry
					if rec:
						for e in _scandir_rec(entry, files, dirs, others, rec, follow_symlinks, errorfunc):
							yield e
				else:
					if others:
						yield entry
	except OSError as e:
		errorfunc(rootentry, e)

def scandir_rec(path, files=True, dirs=False, others=False, rec=True, follow_symlinks=True, relative=False, allow_skip=False, errorfunc=scandir_error_log):
	# type: (PathType, bool, bool, bool, bool, bool, bool, bool, Callable[[MyDirEntryT, Exception], None]) -> Iterator[MyDirEntryT]

	if not logical_implication(allow_skip, dirs and rec):
		raise ValueError("allow_skip implies dirs and rec")

	if isinstance(path, PathLike):
		path = fspath(path)

	entry = DirEntryStub(os.path.basename(path), uncabspath(path)) # for python 2 compat. and long filename support

	if not allow_skip:
		it = _scandir_rec(entry, files, dirs, others, rec, follow_symlinks, errorfunc)
	else:
		it = _scandir_rec_skippable(entry, files, others, follow_symlinks, errorfunc)

	if not relative:
		return it
	else:
		basepath = uncabspath(path)
		if not allow_skip:
			# entry is a `os.DirEntry`
			def modpathrelative(entry):
				entry = MyDirEntry(entry)
				entry.basepath = basepath
				return entry
		else:
			# entry is a `MyDirEntry`
			def modpathrelative(entry):
				entry.basepath = basepath
				return entry

		return map(modpathrelative, it)

def _scandir_depth(rootentry, depth, errorfunc):

	try:
		for entry in scandir(rootentry):
			if entry.is_dir():
				yield depth, entry
				for depth_entry in _scandir_depth(entry, depth + 1, errorfunc):
					yield depth_entry

		for entry in scandir(rootentry):
			if entry.is_file():
				yield depth, entry

	except OSError as e:
		errorfunc(rootentry, e)

def scandir_depth(path, depth=0, onerror=scandir_error_log):
	# type: (str, int, Callable) -> Iterator[Tuple[int, DirEntry]]

	""" Recursive version of `scandir` which yields the current path depth
		along with the DirEntry.
		Directories are returned first, then files.
		The order is arbitrary otherwise.
	"""

	entry = DirEntryStub(os.path.basename(path), uncabspath(path)) # for python 2 compat. and long filename support
	return _scandir_depth(entry, depth, onerror)

def make_readonly(path, stats=None):
	# type: (PathType, Optional[os.stat_result]) -> None

	"""deletes all write flags"""
	if not stats:
		stats = os.stat(path)
	os.chmod(path, stat.S_IMODE(stats.st_mode) & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)

def make_writeable(path, stats=None):
	# type: (PathType, Optional[os.stat_result]) -> None

	"""set owner write flag"""
	if not stats:
		stats = os.stat(path)
	os.chmod(path, stat.S_IMODE(stats.st_mode) | stat.S_IWRITE)

def is_writeable(stats):
	# type: (os.stat_result, ) -> bool

	return stats.st_mode & stat.S_IWRITE

def is_readable(stats):
	# type: (os.stat_result, ) -> bool

	return stats.st_mode & stat.S_IREAD

def isfile(stats):
	# type: (os.stat_result, ) -> bool

	return stat.S_ISREG(stats.st_mode)

def isdir(stats):
	# type: (os.stat_result, ) -> bool

	return stat.S_ISDIR(stats.st_mode)

# was: islink
def statsislink(stats):
	# type: (os.stat_result, ) -> bool

	return stat.S_ISLNK(stats.st_mode)

def extract_basic_stat_info(stats):
	# type: (os.stat_result, ) -> tuple

	""" Returns size, modification, creation time and access mode. """
	return (stats.st_size, stats.st_mtime, stats.st_ctime, stats.st_mode)

def safe_filename(filename, replacement=""):
	# type: (str, str) -> str

	WIN = windows_illegal_chars
	UNIX = unix_illegal_chars
	MAC = mac_illegal_chars
	bad = set.union(WIN, UNIX, MAC)

	safe_filename_translation_table = str.maketrans({c: replacement for c in bad})

	return filename.translate(safe_filename_translation_table) # fixme: return callable which accepts only filename

safe_filename_simple = safe_filename

def _char_subber(s, illegal_chars, replacement):
	# type: (str, Set[str], str) -> str

	if set(replacement) & illegal_chars:
		raise ValueError("replace character cannot be a illegal character")

	regex = "[" + re.escape("".join(illegal_chars)) + "]"
	return re.sub(regex, replacement, s)

def _char_splitter(s, chars):
	# type: (str, Set[str]) -> str

	regex = "[" + re.escape("".join(chars)) + "]"
	return re.split(regex, s)

def _special_subber(s, replacement="_"):
	# type: (str, str) -> str

	if s == "." or s == "..":
		return s.replace(".", replacement)

	return s

def windows_compliant_filename(filename, replacement="_"):
	# type: (str, str) -> str
	""" https://msdn.microsoft.com/en-us/library/aa365247.aspx """
	# reg. wikipedia: \x7f is valid!
	# trailing dots and spaces are ignored by windows

	ret = _char_subber(filename, windows_illegal_chars, replacement).rstrip(". ") # strip or replace?, translate

	root, _ = os.path.splitext(ret)

	if root.upper() in windows_reserved_names:
		raise WindowsIllegalFilename("filename would result in reserved name")

	return ret

def windows_compliant_dirname(filename, replacement):
	# type: (str, str) -> str
	""" https://msdn.microsoft.com/en-us/library/aa365247.aspx """
	# reg. wikipedia: \x7f is valid!
	# trailing dots and spaces are ignored by windows

	ret = _char_subber(filename, windows_illegal_chars, replacement).rstrip(" ")

	root, _ = os.path.splitext(ret)

	if root.upper() in windows_reserved_names:
		raise WindowsIllegalFilename("filename would result in reserved name")

	return ret

def linux_compliant_filename(filename, replacement="_"):
	# type: (str, str) -> str

	filename = _special_subber(filename, replacement)
	return _char_subber(filename, unix_illegal_chars, replacement)

def linux_compliant_dirname(filename, replacement="_"):
	# type: (str, str) -> str

	return _char_subber(filename, unix_illegal_chars, replacement)

def mac_compliant_filename(filename, replacement="_"):
	# type: (str, str) -> str

	filename = _special_subber(filename, replacement)
	return _char_subber(filename, mac_illegal_chars, replacement)

def mac_compliant_dirname(filename, replacement="_"):
	# type: (str, str) -> str

	return _char_subber(filename, mac_illegal_chars, replacement)

def windows_split_path(s):
	# type: (str, ) -> List[str]

	return _char_splitter(s, windows_sep)

def linux_split_path(s):
	# type: (str, ) -> List[str]

	return _char_splitter(s, linux_sep)

def mac_split_path(s):
	# type: (str, ) -> List[str]

	return _char_splitter(s, mac_sep)

system = platform.system()

if system == "Windows":
	compliant_filename = windows_compliant_filename
	compliant_dirname = windows_compliant_dirname
	split_path = windows_split_path
elif system == "Linux":
	compliant_filename = linux_compliant_filename
	compliant_dirname = linux_compliant_dirname
	split_path = linux_split_path
elif system == "Darwin":
	compliant_filename = mac_compliant_filename
	compliant_dirname = mac_compliant_dirname
	split_path = mac_split_path
else:
	compliant_filename = _not_available("compliant_filename")
	compliant_dirname = _not_available("compliant_dirname")
	split_path = _not_available("split_path")

def compliant_path(path, replace="_"):
	# type: (str, str) -> str

	fn_func = partial(compliant_dirname, replace=replace)
	path_compontents = split_path(path)

	ret = os.sep.join(map(fn_func, path_compontents[:-1]))
	ret = os.path.join(ret, compliant_filename(path_compontents[-1]))
	return ret

def reldirname(path):
	# type: (PathType, ) -> str

	ret = os.path.dirname(path)
	if not ret:
		ret = "."

	return ret

def equal_dirs_iter(dir1, dir2, follow_symlinks=False):

	""" Tests if two directories are equal. Doesn't handle links. """

	def entry_path_size(entry):
		return entry.path, entry.stat().st_size

	files1 = sorted(map(entry_path_size, scandir_rec(dir1, follow_symlinks=follow_symlinks))) # get rel paths
	files2 = sorted(map(entry_path_size, scandir_rec(dir2, follow_symlinks=follow_symlinks))) # get rel paths

	for (path1, size1), (path2, size2) in zip_longest(files1, files2):
		if path1 != path2:
			yield ("name", path1, path2)

		elif size1 != size2:
			yield ("size", path1, path2)

	for (path1, size1), (path2, size2) in zip_longest(files1, files2):
		if path1 == path2 and size1 == size2 and not equal_files(path1, path2):
			yield ("data", path1, path2)

def equal_dirs(dir1, dir2):
	return is_empty(equal_dirs_iter(dir1, dir2))

def realpath_win(path):
	# type: (PathType, ) -> str

	"""fix for os.path.realpath() which doesn't work under Win7"""

	path = os.path.abspath(path)
	if islink(path):
		return os.path.normpath(os.path.join(os.path.dirname(path), os.readlink(path)))
	else:
		return path

def search(directories, pattern, dirs=True, files=True, rec=True, follow_symlinks=False):
	# type: (Iterable[PathType], str, bool, bool, bool, bool) -> Iterator[DirEntry]

	""" Search for files and folders matching the wildcard `pattern`. """

	for directory in directories:
		for entry in scandir_rec(directory, dirs=dirs, files=files, rec=rec, follow_symlinks=follow_symlinks):
			if fnmatch(entry.name, pattern):
				yield entry

def rename_files_in_folder(path, pattern, transform, template, rec=True, follow_symlinks=False):
	# type: (PathType, str, Callable, str, bool, bool) -> None

	cpattern = re.compile(pattern)

	for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
		filepath = entry.path
		base, name = os.path.split(filepath)

		m = cpattern.match(name)
		if not m:
			continue
		args = transform(*m.groups())

		to = template.format(*args)
		os.rename(filepath, os.path.join(base, to))

def clean_directory_by_extension(path, ext, rec=True, follow_symlinks=False):
	# type: (PathType, str, bool, bool) -> None

	""" Deletes all files of type `ext` if the same file without the ext exists
		and rename if it does not exit. """

	#delete .ext if normal file exists
	for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
		filepath = entry.path
		if not filepath.endswith(ext):
			file_to_delete = filepath + ext
			if os.path.isfile(file_to_delete):
				logger.info("Remove: %s", file_to_delete)
				try:
					os.remove(file_to_delete)
				except OSError as e:
					logger.exception("Could not delete: %s", file_to_delete)
					break

	#rename all .ext to normal
	for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
		filepath = entry.path
		if entry.name.endswith(ext):
			target = filepath[:-len(ext)]
			try:
				logger.info("Rename: %s -> %s", filepath, target)
				os.rename(filepath, target)
			except OSError as e:
				logger.exception("Should not have happened")
				break

def iter_links(path):
	# type: (PathType, ) -> Iterator[str]

	""" Yields all directory symlinks (and junctions on windows). """

	for dirpath, dirnames, filenames in os.walk(path):
		for dirname in dirnames:
			joined = os.path.join(dirpath, dirname)
			if islink(joined):
				yield joined

def normalize_seps(path):
	# type: (str, ) -> str

	""" Converts \ to /
	"""

	return path.replace("\\", "/")

def entrysuffix(entry):
	# type: (MyDirEntryT, ) -> str

	return os.path.splitext(entry.name)[1]

class Counts(object):
	__slots__ = ("dirs", "files", "others")

	def __init__(self):
		self.dirs = 0
		self.files = 0
		self.others = 0

	def __iadd__(self, other):
		# type: (Counts, ) -> None

		self.dirs += other.dirs
		self.files += other.files
		self.others += other.others

	def null(self):
		# type: () -> bool

		return self.dirs == 0 and self.files == 0 and self.others == 0

def _scandir_counts(rootentry, files=True, others=True, rec=True, total=False, errorfunc=scandir_error_raise):
	# type: (MyDirEntryT, bool, bool, bool, bool, Callable) -> Iterator[DirEntry, Optional[Counts]]

	counts = Counts()

	try:
		for entry in scandir(rootentry.path):

			if entry.is_dir():
				counts.dirs += 1
				if rec:
					for subentry, subcounts in _scandir_counts(entry, files, others, rec, total, errorfunc):
						yield subentry, subcounts

						if total:
							counts += subcounts

			elif entry.is_file():
				counts.files += 1
				if files:
					yield entry, None

			else:
				counts.others += 1
				if others:
					yield entry, None

		yield rootentry, counts  # yield after the loop

	except OSError as e:
		errorfunc(rootentry, e)

def scandir_counts(path, files=True, others=True, rec=True, total=False, onerror=scandir_error_log):
	# type: (PathType, bool, bool, bool, bool, Callable) -> Iterator[DirEntry, Optional[Counts]]

	""" A recursive variant of scandir() which also returns the number of files/directories
		within directories.
		If total is True, the numbers will be calculated recursively as well.
	"""

	entry = DirEntryStub(os.path.basename(path), uncabspath(path)) # for python 2 compat. and long filename support
	return _scandir_counts(entry, files, others, rec, total, onerror)

def shutil_onerror_remove_readonly(func, path, exc_info):

	""" onerror function for the `shutil.rmtree` function.
		Makes read-only files writable and tries again.
	"""

	stats = os.stat(path)
	if is_readable(stats):
		make_writeable(path, stats)
		func(path)
	else:
		raise



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range
from copy import deepcopy
from random import choice
from typing import TYPE_CHECKING

from .compat.math import isqrt
from .exceptions import assert_type
from .indexing import row_indices, col_indices, subblock_indices
from .iter import batch
from .set import get as setget

if TYPE_CHECKING:
	from typing import Iterable, Set, TypeVar
	T = TypeVar("T")

class Unsolvable(Exception):
	pass


class Sudoku(object):

	def __init__(self, board, sym_set, sym_free):
		# type: (Iterable[T], Set[T], T) -> None

		assert_type("sym_set", sym_set, set)

		self.square = board
		self.sym_set = sym_set
		self.sym_free = sym_free

		self.outer_square_size = isqrt(len(board))
		self.outer_square_area = self.outer_square_size ** 2
		if self.outer_square_area != len(board):
			raise ValueError("board has an invalid size")

		if self.outer_square_size != len(sym_set):
			raise ValueError("sym_set length must be equal to the edge length of the board: {} vs {}".format(len(sym_set), self.outer_square_area))

		self.inner_square_size = isqrt(self.outer_square_size)
		if self.inner_square_size ** 2 != self.outer_square_size:
			raise ValueError("board has an invalid size")

		self.solved = False

	def init_board(self, board):
		raise NotImplementedError

	def get_board(self):
		raise NotImplementedError

	def print_square(self):
		for i, num in enumerate(self.get_board(), 1):
			print(num, end=" ")
			if i % self.outer_square_size == 0:
				print()

	def solve(self, strategy=None):
		self.square = self.init_board(self.square)
		return self._solve(strategy)


class SudokuRulebased(Sudoku):

	def init_board(self, board):
		square = []
		for i in board:
			if i != self.sym_free:
				square.append({i})
			else:
				square.append(self.sym_set.copy())
		return square

	def get_board(self):
		for num in self.square:
			if len(num) == 1:
				yield setget(num)
			else:
				yield num

	def get_row_nums(self, i):
		row_set = set()
		for x in row_indices(i, self.outer_square_size):
			if len(self.square[x]) == 1 and x != i:
				row_set.update(self.square[x])
		return row_set

	def get_column_nums(self, i):
		column_set = set()
		for x in col_indices(i, self.outer_square_size):
			if len(self.square[x]) == 1 and x != i:
				column_set.update(self.square[x])
		return column_set

	def get_inner_square_nums(self, n):
		inner_square_set = set()
		for k in subblock_indices(n, self.outer_square_size, self.inner_square_size):
			if len(self.square[k]) == 1 and k != n:
				inner_square_set.update(self.square[k])
		return inner_square_set

	def update_cell(self, i):
		a = self.get_row_nums(i)
		b = self.get_column_nums(i)
		c = self.get_inner_square_nums(i)
		self.square[i] -= a
		self.square[i] -= b
		self.square[i] -= c

	def _solve(self, strategy=None):
		while not self.solved:
			self.solved = True
			square_old = deepcopy(self.square)
			for i in range(self.outer_square_area):
				if len(self.square[i]) != 1:
					self.solved = False
					#print_confirm("#%u is unsolved -> solve" % i)
					self.update_cell(i)
					#print_confirm(self.square)
			if square_old == self.square and not self.solved:
				raise NotImplementedError("Sudoku not solvable using this method")
		return self.square


class SudokuBruteforce(Sudoku):

	def init_board(self, board):
		return board

	def get_board(self):
		return self.square

	def get_row_nums(self, i):
		row_set = set()
		for x in row_indices(i, self.outer_square_size):
			if x != i:
				row_set.add(self.square[x])
		return row_set

	def get_column_nums(self, i):
		column_set = set()
		for x in col_indices(i, self.outer_square_size):
			if x != i:
				column_set.add(self.square[x])
		return column_set

	def get_inner_square_nums(self, i):
		inner_square_set = set()
		for x in subblock_indices(i, self.outer_square_size, self.inner_square_size):
			if x != i:
				inner_square_set.add(self.square[x])
		return inner_square_set

	def get_possible_nums(self, i):
		# type: (int, ) -> Set[T]

		all = self.get_row_nums(i) | self.get_column_nums(i) | self.get_inner_square_nums(i)
		return self.sym_set - all

	def get_next_higher_possible_num(self, i, num):
		for pn in sorted(self.get_possible_nums(i)): # fixme: sorting is bad here. # why? for algorithmic performance or python performance?
			if pn > num:
				return pn

		return self.sym_free

	def get_next_lower_possible_num(self, i, num):
		for pn in sorted(self.get_possible_nums(i), reverse=False): # see: get_next_higher_possible_num
			if pn < num:
				return pn

		return self.sym_free

	def _solve(self, strategy=None):

		strategy = strategy or "inc"

		try:
			candidate_func = {
				"inc": self.get_next_higher_possible_num,
				"dec": self.get_next_lower_possible_num,
			}[strategy]
			
			selection_func = {
				"inc": min,
				"dec": max,
			}[strategy]

		except KeyError:
			raise ValueError("Invalid strategy")

		i = 0
		backtrack = []  # type: Tuple[int, T]
		backtracks = 0
		steps = 0

		while i < self.outer_square_area:
			if self.square[i] != self.sym_free:
				i+=1
				continue
			pnums = self.get_possible_nums(i)

			if len(pnums) == 0:
				while True:
					backtracks += 1
					try:
						j, num = backtrack.pop()
					except IndexError:
						raise Unsolvable

					num = candidate_func(j, num)
					self.square[j] = num
					if num != self.sym_free:
						backtrack.append((j, num))
						i = j + 1
						break
			else:
				num = selection_func(pnums)
				self.square[i] = num
				backtrack.append((i, num))
				steps += 1

		return backtracks, steps


def is_valid_solution(board, sym_set):
	# type: (Iterable[T], Set[T]) -> bool

	""" Checks if `board` is a valid solved Sudoku configuration
	"""

	edge_len = isqrt(len(board))
	square_len = isqrt(edge_len)

	board = tuple(batch(board, edge_len, func=tuple))

	# check rows
	for j in range(edge_len):
		row = set(board[j][i] for i in range(edge_len))
		if set(row) != sym_set:
			return False

	# check cols
	for i in range(edge_len):
		col = set(board[j][i] for j in range(edge_len))
		if col != sym_set:
			return False

	# check inner squares
	for x in range(0, edge_len, square_len):
		for y in range(0, edge_len, square_len):
			square = set(board[i][j] for i in range(x, x + square_len) for j in range(y, y + square_len))
			if square != sym_set:
				return False

	return True


if __name__ == "__main__":

	from argparse import ArgumentParser
	from genutility.time import MeasureTime

	parser = ArgumentParser()
	parser.add_argument("board", metavar="N", type=int, nargs="+", help="Flat Sudoku board")
	parser.add_argument("--symbols", default=set(range(1, 10)), type=int, nargs="+", help="Used symbols")
	parser.add_argument("--free", default=0, help="Free marker symbol")
	parser.add_argument("--strategy", choices=("inc", "dec"), default="linear")
	args = parser.parse_args()

	sym_set = set(args.symbols)

	s = SudokuBruteforce(args.board, sym_set, args.free)

	try:
		with MeasureTime() as t:
			steps, backtracks = s.solve(args.strategy)

		s.print_square()
		if is_valid_solution(args.board, sym_set):
			print("Solved sudoku in {:.2} seconds, using {} steps and backtracked {} times.".format(t.get(), steps, backtracks))
		else:
			print("No valid solution was found in {:.2} seconds, using {} steps and backtracked {} times.".format(t.get(), steps, backtracks))
	except Unsolvable:
		print("Sudoku is not solvable")



from __future__ import absolute_import, division, print_function, unicode_literals

try:
	from sys import hash_info

	def get_wordsize():
		return hash_info.width

except ImportError:
	import struct
	wordsize = struct.calcsize("P") * 8

	def get_wordsize():
		return wordsize

__all__ = ["get_wordsize"]



import tensorflow as tf

def categorical(probs, size=1):
	logits = tf.math.log(probs)[None,:]
	return tf.squeeze(tf.random.categorical(logits, num_samples=size), axis=0)

def randint(high, size=(), dtype=tf.int64):
	return tf.random.uniform(size, maxval=high, dtype=dtype)



from __future__ import absolute_import, division, print_function, unicode_literals

# from builtins import str # don't do this here because of cookie construction!
from future.utils import iteritems, PY2
from http.cookies import SimpleCookie as IncompatibleSimpleCookie

if PY2:
	class SimpleCookie(IncompatibleSimpleCookie):

		def __init__(self, rawdata):
			if isinstance(rawdata, dict):
				rawdata = {str(k): v for k, v in iteritems(rawdata)}
			elif isinstance(rawdata, unicode): # noqa: F821
				rawdata = str(rawdata)

			IncompatibleSimpleCookie.__init__(self, rawdata)
else:
	SimpleCookie = IncompatibleSimpleCookie



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewkeys, viewvalues
from itertools import chain, repeat
from typing import TYPE_CHECKING

from .exceptions import InconsistentState, NoResult

if TYPE_CHECKING:
	from typing import Any, Iterator
	from .typing import Connection, Cursor

class TransactionCursor(object):

	""" Cursor context manager which starts a transaction and rolls back in case of error.
	"""

	def __init__(self, conn):
		# type: (Connection, ) -> None

		self.cursor = conn.cursor()

	def __enter__(self):
		# type: () -> Cursor

		self.cursor.execute("BEGIN TRANSACTION")
		return self.cursor

	def __exit__(self, exc_type, exc_value, traceback):
		# type: (Any, Any, Any) -> None

		if exc_type:
			self.cursor.execute("ROLLBACK TRANSACTION")
		else:
			self.cursor.execute("COMMIT TRANSACTION")
		self.cursor.close()

class CursorContext(object):

	""" Cursor context manager which creates a new cursor and closes it when it leaves the context.
	"""

	def __init__(self, conn):
		# type: (Connection, ) -> None

		self.cursor = conn.cursor()

	def __enter__(self):
		# type: () -> Cursor

		return self.cursor

	def __exit__(self, exc_type, exc_value, traceback):
		# type: (Any, Any, Any) -> None

		self.cursor.close()

def upsert(cursor, primary, values, table):
	# type: (Cursor, dict, dict, str) -> bool

	""" Inserts `table` fields specified by `primary` and `values` keys,
		with the corresponding values.
		If all the values specified by `primary` already exist,
		they are updated instead.

		`table`, `primary.keys()` and `values.keys()` are not escaped.
		`primary.values()` and `values.values()` are escaped.
	"""

	# use INSERT ... ON DUPLICATE KEY UPDATE instead?

	if not primary:
		raise ValueError("Empty primary mapping would result in an empty WHERE condition which would affect all rows")

	set_str = ",".join("{}=?".format(k) for k in viewkeys(values))
	where_str = " AND ".join("{}=?".format(k) for k in viewkeys(primary))

	cursor.execute("UPDATE {} SET {} WHERE {}".format(table, set_str, where_str),  # nosec
		chain(viewvalues(values), viewvalues(primary))
	)

	if cursor.rowcount == 0:
		into_str = ",".join(chain(viewkeys(primary), viewkeys(values)))
		values_str = ",".join(repeat("?", len(primary) + len(values)))
		cursor.execute("INSERT INTO {} ({}) VALUES ({})".format(table, into_str, values_str),  # nosec
			chain(viewvalues(primary), viewvalues(values))
		)
		return True

	return False

def fetchone(cursor):
	# type: (Cursor, ) -> Any

	""" Fetch results from `cursor` and assure only one result was returned.
	"""

	rows = cursor.fetchall()
	if len(rows) == 0:
		raise NoResult("No result found")
	elif len(rows) == 1:
		return rows[0]
	else:
		raise InconsistentState("More than one result found")

def iterfetch(cursor, batchsize=1000):
	# type: (Cursor, int) -> Iterator[Any]

	""" Iterate all results from `cursor`.
	"""

	while True:
		results = cursor.fetchmany(batchsize)
		if not results:
			break
		for result in results:
			yield result



from __future__ import absolute_import, division, print_function, unicode_literals

import logging, os.path
from sys import stdout
from functools import wraps, reduce, partial
from time import sleep
from typing import TYPE_CHECKING

from .iter import retrier

if TYPE_CHECKING:
	from typing import Callable, Any, TextIO, Iterable, Iterator, Optional, TypeVar, Tuple, Sequence
	T = TypeVar("T")
	U = TypeVar("U")

logger = logging.getLogger(__name__)

class NotRetried(RuntimeError):
	pass

def identity(x):
	# type: (T, ) -> T

	""" Identity function. """

	return x

def nop():
	# type: () -> None

	""" Function which does absolutely nothing (aka pass, noop). """

	pass

def partial_decorator(*args, **kwargs):
	# type: (*Any, **Any) -> Callable

	""" Same as `functools.partial` but applied as a decorator. """

	def decorator(func):
		return partial(func, *args, **kwargs)
	return decorator

def compose_two(f, g):
	# type: (Callable, Callable) -> Callable

	""" compose_two(f, g) -> lambda x: f(g(x)) """

	return lambda x: f(g(x))

def compose(*functions):
	# type: (Iterable[Callable[[Any], Any]], ) -> Callable[[Any], Any]

	""" compose(f, g, h) -> lambda x: f(g(h(x))).
		see: Function composition
	"""

	return reduce(compose_two, functions, identity)

def apply(f, x): # *args, **kwargs
	# type: (Callable[[T], U], T) -> U

	""" Same as `f(x)`. """

	return f(x)

def zipmap(funcs, vals):
	# type: (Iterable[Callable[[T], U]], Iterable[T]) -> Iterator[U]

	""" Applies a list of functions to a list of values. """

	return map(apply, funcs, vals)

def multiapply(funcs, elm):
	# type: (Iterable[Callable], Any) -> Any

	""" Applies functions `funcs` to element `elm` iteratively. """

	for func in funcs:
		elm = func(elm)

	return elm

def multimap(funcs, it):
	# type: (Sequence[Callable], Iterable[Any]) -> Iterator[Any]

	""" Applies functions `funcs` to each element of `it` iteratively. """

	for i in it:
		yield multiapply(funcs, i)

def call_repeated(num):
	# type: (int, ) -> Callable[[Callable], Callable]

	""" Function decorator to call decorated function `num` times with the same arguments.
		Returns the results of the last call.
	"""

	if num < 1:
		raise ValueError("num must be larger than 0")

	def dec(func):
		@wraps(func)
		def inner(*args, **kwargs):
			last = None
			for i in range(num):
				last = func(*args, **kwargs)
			return last
		return inner
	return dec

def print_return_type(func, file=stdout):
	# type: (Callable, TextIO) -> Callable

	""" Wraps function to print the return type after calling. """

	@wraps(func)
	def inner(*args, **kwargs):
		ret = func(*args, **kwargs)
		print(type(ret), file=file)
		return ret
	return inner

def retry(func, waittime, exceptions=(Exception, ), attempts=-1, multiplier=1, jitter=0, max_wait=None, jitter_dist="uniform", waitfunc=sleep):
	# type: (Callable[[], T], float, Tuple[Exception], int, float, float, Optional[float], str, Callable[[float], Any]) -> T

	""" Retry function `func` multiple times in case of raised `exceptions`.
		See `genutility.iter.retrier()` for the remaining arguments.
		Reraises the last exception in case the function call doesn't succeed after retrying.
	"""

	last_exception = None # type: Optional[Exception]
	for i in retrier(waittime, attempts, multiplier, jitter, max_wait, jitter_dist, waitfunc):
		try:
			return func()
		except exceptions as e:
			logger.info("Attempt %s failed: %s", i+1, e)
			last_exception = e

	if last_exception:
		raise last_exception  #pylint: disable=raising-bad-type
	else:
		raise NotRetried

class CustomCache(object):

	""" Class to build decorator cache function using custom reader and writer functions.
		The cache method ignores all arguments of the decorated function.

		Example:
		```
		cc = CustomCache(read_pickle, write_pickle)

		@cc.cache("func-cache.p")
		# arg is ignored, so the cache will return the result of the argument which was supplied
		# when the cache file was created.
		def func(arg): 
			return arg
		```
		a = func(1) # cache created, a == 1
		b = func(2) # cache loaded, b == 1
	"""

	def __init__(self, reader, writer):
		# type: (Callable, Callable) -> None

		self.reader = reader
		self.writer = writer

	def cache(self, path):
		# type: (str, ) -> Callable

		def dec(func):
			# type: (Callable, ) -> Callable

			@wraps(func)
			def inner(*args, **kwargs):
				# type: (*Any, **Any) -> Any

				if os.path.exists(path):
					logger.debug("Loading object from cache %s", path)
					return self.reader(path)
				else:
					logger.debug("Saving obect to cache %s", path)
					result = func(*args, **kwargs)
					self.writer(result, path)
					return result

			return inner
		return dec



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from pandas import DataFrame, Series

from .func import identity

if TYPE_CHECKING:
	from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar
	T = TypeVar("T")

def pandas_json(obj):
	# type: (Any, ) -> Any

	""" Can be used for the `json.dump` `default` argument
		to make some pandas objects JSON serializable.
	"""

	if isinstance(obj, DataFrame):
		return dict(obj)
	elif isinstance(obj, Series):
		return tuple(obj)

	raise TypeError("object of type {} cannot be JSON serialized: {}".format(type(obj), obj))

def strlist(sep):
	# type: (str, ) -> Callable[[str], List[str]]

	def inner(s):
		if s:
			return s.split(sep)
		else:
			return []

	return inner

def dataframe_to_dict(df, empty=False, cellfunc=None, dictcls=dict):
	# type: (DataFrame, bool, Optional[Callable], Callable[[Iterable[Tuple[str, Any]]], T]) -> T

	""" Turns dataframes into trees. Columns first, then rows
	"""

	if cellfunc is None:
		cellfunc = identity

	def cols(series):
		for rowname, rowdata in series.iteritems():
			value = cellfunc(rowdata)
			if value or empty:
				yield rowname, value

	def rows(df):
		for colname, coldata in df.iteritems():
			value = dictcls(cols(coldata))
			if value or empty:
				yield colname, value

	return dictcls(rows(df))
